/*
** Module   : Utils.java
** Abstract : General purpose utility methods
**
** Copyright (c) 2004-2023, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------------Description-----------------------------------
** 001 ECF 20041122   @18804 Created initial version, consisting of some
**                           Properties object helper methods.
** 002 NVS 20050223   @19935 Added readPassphrase method.
** 003 ECF 20050225   @19962 Added printChainedErrors method.
** 004 ECF 20050225   @20061 Added findFileInPath method. Generalized and
**                           moved from findFile in uast.AstGenerator.
** 005 GES 20050304   @20196 Added file copying methods.
** 006 GES 20050311   @20293 Added isLaterVersion method.
** 007 ECF 20050329   @20519 Modified printChainedErrors method. Now
**                           prints name of throwable type if no message
**                           is present.
** 008 ECF 20050422   @20844 Added indent parameter to printChainedErrors
**                           method. Indents each message line by the
**                           specified number of spaces.
** 009 GES 20050706   @21667 Added append mode to copyFile().
** 010 ECF 20051103   @23222 Added getLog method. Retrieves an Apache
**                           commons-logging logging object scoped to a
**                           particular package.
** 011 ECF 20051104   @23239 Added describeContext method. Uniquely
**                           describes current context for debugging or
**                           logging.
** 012 ECF 20060112   @23897 Added verifyDirectoryNode method. Verifies
**                           existence and class of a required directory
**                           entry.
** 013 GES 20060122   @24013 Added directory query helpers to manage
**                           the entire process of searching for a
**                           specified value across a pre-defined
**                           hierarchy in the directory.
** 014 ECF 20060123   @24039 Modified describeContext method. Combine
**                           session ID, thread ID, and user ID to create
**                           a unique string.
** 015 GES 20060125   @24073 Added directory node search helper.
** 016 NVS 20060226   @24749 Added generic method search and invocation
**                           service invoke().
** 017 NVS 20060227   @24771 Added noHelp() method which is a placeholder
**                           for a real help application.
** 018 ECF 20060309   @24998 Added getUnqualifiedName() method. Gets the
**                           short name of a class or interface.
** 019 GES 20060609   @27053 Removed fillInStackTrace() from invoke()
**                           since this only destroys the exception's
**                           stack trace (it doesn't affect any chained
**                           cause's stack trace). It is better to have
**                           the original record of where the problem
**                           occurred than to know that we rethrew the
**                           exception from this point.
** 020 GES 20060802   @28296 Created primitive/wrapper conversion helpers
**                           to simplify list and array processing.
** 021 ECF 20060918   @29668 Temporary fix to describeContext(). Handle
**                           invalid thread ID returned from
**                           SecurityManager. A better fix would be to
**                           ensure SecurityManager never returns a null
**                           thread ID.
** 022 GES 20061220   @31807 Removed readPassphrase() and replaced it
**                           with prompt() that is a simpler and safer
**                           alternative.
** 023 NVS 20070112   @31893 Added uniqueId() method and supporting data
**                           in form of a context local work area.
**                           Added methods to work with sets: intersects()
**                           intersection() and difference().
** 024 NVS 20070126   @32010 Added scope param to findDirectoryNodePath()
**                           method. Now the method can do server level
**                           searches (/server/<serverId> and
**                           /server/default paths) and account level
**                           searches as the original implementation.
** 025 NVS 20070404   @32764 Added new method diffIntArrays().
** 026 GES 20070716   @34530 Moved file copying related methods here for
**                           better reuse.
** 027 ECF 20071009   @35600 Added getPortForService() and variants of
**                           other directory convenience methods. The
**                           former looks up the port number for a given
**                           service name. The latter add a 'scope'
**                           parameter to permit looking up values of each
**                           basic data type by account/group or by
**                           server.
** 028 ECF 20071101   @35897 Changed invoke() signature. Now uses var-args
**                           instead of an Object array for parameters, if
**                           any, to be passed to the invoked method.
** 029 GES 20080414   @38023 Added helpers for text processing which
**                           honors the proper charset conversions.
** 030 GES 20080421   @38058 Added another charset helper.
** 031 CA  20081006   @40061 Added integerCollectionToPrimitive.
** 032 CA  20081205   @40820 Added integerMapToPrimitive.
** 033 CA  20090413   @41770 Fixed bug in integerMapToPrimitive - index was
**                           not incremented.
** 034 GES 20090422   @41918 Converted to standard string formatting.
** 035 GES 20090424   @41975 Import change.
** 036 NVS 20090624   @42833 Added new method removeDirectoryNode().
** 037 SIY 20090701   @43013 Added scroll() method. 
**                           Removed integerListToPrimitive() which is a copy
**                           of integerCollectionToPrimitive(). Inferred 
**                           generics in order to improve type checking.
** 038 SIY 20090923   @43987 Directory binding in getDirectoryNodeWorker now is
**                           optional if directory is already bound.  
** 039 ECF 20101004          Added methods to match a file path case-
**                           insensitively and to canonicalize a file system
**                           path.
** 040 SIY 20110112          Changes related to new color handling scheme. Minor
**                           cleanups.
** 041 GES 20110623          Fixed case sensitive filename matching helpers.
** 042 CA  20130822          Added API to check if two sets match.
** 043 HC  20131031          Support methods for Long resource IDs, see issue #2183.
** 044 CA  20140206          Improved the directory node lookup APIs, to allow search using an 
**                           explicit account list and to allow fallback to a per-server value, if
**                           per-account search failed.
** 045 OM  20140706          Fixed matchPathCaseInsensitively for Windows filesystem.
** 046 GES 20141106          Improved some javadocs and added unimplemented feature logging.
** 047 CA  20150302          Add logging if a directory node is found with an unexpected class.
** 048 ECF 20150715          Replace StringBuffer with StringBuilder. Deprecated getLog.
** 049 GES 20150923          Added causeChainContains().
** 050 OM  20151130          Eliminated IOException from canonicalizePath. Declared getLog()
**                           method as @Deprecated because we are moving towards J2SE logging.
** 051 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 052 CA  20151229          If instantiation of entry point fails with an ERROR condition, then 
**                           show the error message.
** 053 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 054 IAS 20160229          Added file extension extractor, tryWhileNull utility method and
**                           InstrumentedMap.
** 055 ECF 20160202          Removed deprecated getLog method.
** 056 IAS 20160331          Fixed intersect() and isSame() to avoid NPE.
** 057 OM  20160422          Added per-project support.
** 058 EVL 20170831          Adding helper method to convert string to file size integer value.
** 059 ECF 20171121          Modified intersect method to exit quickly if sets are empty.
**     GES 20171122          Added split() to more efficiently calculate both the differences
**                           and intersection, when both are needed anyway.
** 060 SBI 20180410          Added calculateResourcePath to be used by gui drivers.
** 061 SBI 20180521          Added getPathResolver and getSafePathResolver.
** 062 SBI 20180531          Added common Utils.getDirectoryNodeMap.
** 063 SBI 20180716          Generalized code normalizeZorder and Zkey, moveToTopInClass and
**                           moveToBottomInClass from AbstractContainer.
** 064 HC  20180813          Added double data type to the directory access interface.
** 065 HC  20190701          Added methods longToBytesLE and bytesToLongLE.
** 066 CA  20190731          Added findFieldWithValue.
**     CA  20190811          Added collectInterfaces.
** 067 ECF 20190619          Refactored replaceSeparator from canonicalizePath.
** 068 CA  20191119          Track the instantiating class, to be able to register any buffer to 
**                           its source definition class.
**     CA  20191212          Added findConstructor, to find the first ctor which matches the 
**                           argument's class or super-class, or any of its super-interfaces.
** 069 HC  20200130          Added a method for enumerating nodes.
** 070 SBI 20200225          Moved isValidValue to this helper class.
** 071 CA  20200122          Javadoc fixes.
** 072 HC  20200726          Added enumDirectoryNodes method.
**     GES 20200922          Minor optimizations in set processing.
**     CA  20200924          Added Utils.invoke, to execute methods via ReflectASM instead of Method.invoke.
**     CA  20201003          Use an identity HashSet where possible.
**     CA  20201011          Added intersect() and containsAll() for RoaringBitmap (fast-access bitmap used
**                           instead of integer sets).
**     ME  20210121          Fix longToBytesLE to use little end order.
**     RFB 20210331          Added Utils.onlyDigits to assist FillIn widget updates. Ref #5210. 
**     RFB 20210402          No need for onlyDigits, as NumberType.parseDoubleTest was added to handle this
**                           need. Minor javadoc fix, as well. Ref #5210.
**     SBI 20210413          Added generic getDirectoryNodeBeanList to get simple java bean from the directory settings.
**         20210414          Added generic searchResourceJars, moved getRefinedPathName from FileSystemOps.
**     CA  20210407          Throw NPE if if invoking an instance method and the instance is null.
**     SBI 20210621          Changed getDirectoryNodeBeanList to get properties from node's attributes.
**         20210701          Added the common code to getOrCreateDefaultFwdTemporaryDirectory(), SupplierWithMemo.
**     HC  20210727          Fixed doWithBoundDS to bind the supplied directory service if needed.
**     RFB 20210805          Implemented safe isServer/isClient/isChUI/isGUI for use on client of server. 
**                           Ref #4984.
**     SBI 20210813          Changed searchResourceJars to make it independent from the server side,
**                           added getJarPath
**     ECF 20210916          describeContext now returns only the thread name if not running on the server.
**     CA  20210928          Added getOrCreateTemporaryDirectory.
**     CA  20211004          Automatically delete the created temporary folder (and its content), on JVM
**                           shutdown.
**     CA  20211019          The temporary folder name must be unique for each context (previously was a 
**                           static variable).
**     VVT 20211020          getOrCreateTemporaryDirectory() fixed: cleanup hook must be added only once per directory.
**                           (ref. #5744).
**     VVT 20211021          Restored statements unnecessarily nested within else clause, removed
**                           earlier in the previous change.
**     CA  20211025          Added dfsClassHierarchy.
**     HC  20220307          Added disableClientSSLCertificateValidation to allow disabling of SSL
**                           certificate validation on HTTPS client connections.
**     SBI 20220427          Added int convertCodepoint(int, String), getCodepoint(int, String) and
**                           changed char toChar(int) to use cpinternal charset.
**     RFB 20220504          Minor javadoc fixes.
**     SBI 20220504          Fixed toChar(int, String) to return -1 if key code is not represented by
**                           the given char set.
**     TJD 20220504          Java 11 compatibility minor changes
** 073 SBI 20230224          Added getCharsetOverride(String charsetName).
** 074 SBI 20230411          Added isEndPointAvailable.
** 075 GBB 20230512          Logging methods replaced by CentralLogger/ConversionStatus.
** 076 CA  20230601          Changed 'collectInterfaces' to perform a BFS walk of the type hierarchy.  Also,
**                           this fixes a bug where only the direct interfaces were collected.
** 077 GBB 20230608          pollAll method added to empty a Queue into a List.
** 078 GBB 20230620          createThreadName method added (originally defined in LogHelper).
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package src.com.goldencode.p2j.util;

import src.com.goldencode.p2j.*;
import src.com.goldencode.p2j.util.logging.*;
import src.com.goldencode.util.*;

import javax.net.ssl.*;
import java.io.*;
import java.lang.invoke.*;
import java.lang.reflect.*;
import java.net.*;
import java.nio.*;
import java.nio.charset.*;
import java.nio.file.FileSystem;
import java.nio.file.*;
import java.security.*;
import java.security.cert.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.function.*;
import java.util.logging.*;
import java.util.stream.*;

/**
 * Collection of useful, static, utility methods for various tasks.
 */
public final class Utils
{
   /** Define search modes for the directory node lookups. */
   public static enum DirScope
   {
      SERVER,
      ACCOUNTS,
      BOTH
   }
   
   /** logger */
   private static final CentralLogger LOG = CentralLogger.get(Utils.class.getName());

   /** Charset name to use for encoding or <code>null</code> for default. */
   private static String override = null;
   
   /** Tracks if the static data has been initialized. */
   private static boolean init = false;
   
   /** Logging token that ensures unimplemented feature logging occurs only once. */
   private static Object token = null;
   
   /** A cache of single-arg constructors in all possible sub-classes. */
   private static Map<String, Constructor<?>> bdtConstructors;
   
   /**
    * Private constructor to prevent instances of this class from being
    * created.
    */
   private Utils()
   {
   }
   
   /**
    * Find the first single-argument constructor which has a single parameter with the argument's
    * type or any of its super-classes or super-interfaces.
    * 
    * @param    cls
    *           The class where to search for the c'tor.
    * @param    arg
    *           The argument's type.
    *           
    * @return   The found c'tor.
    */
   public static Constructor<?> findConstructor(Class<?> cls, Class<?> arg)
   {
      String key = cls.toString() + "#";

      Class<?> parent = arg;
      while (parent != null)
      {
         Constructor<?> ctor = bdtConstructors.get(key + parent.toString());
         if (ctor != null)
         {
            return ctor;
         }
         
         parent = parent.getSuperclass();
      }
      
      // look for the interfaces
      Set<Class<?>> ifaces = Collections.newSetFromMap(new IdentityHashMap<>());
      collectInterfaces(cls, ifaces);
      for (Class<?> ifc : ifaces)
      {
         Constructor<?> ctor = bdtConstructors.get(key + ifc.toString());
         if (ctor != null)
         {
            return ctor;
         }
      }
      
      return null;
   }

   /**
    * Scroll given array so element with specified index will be first in
    * array. All elements before it will be placed at the end of the array.
    * <p>
    * NOTE: Scrolling is performed "in place", so input array gets updated!
    * 
    * @param    <T>
    * @param    source
    *           Source array to scroll.
    * @param    first
    *           The index of element which will be first.
    *           
    * @return   reference to the source array.
    */
   public static<T> T[] scroll(T[] source, int first)
   {
       if (first <= 0 || first >= source.length - 1)
           return source;
       
       T[] tmp1 = Arrays.copyOfRange(source, 0, first);
       T[] tmp2 = Arrays.copyOfRange(source, first, source.length);
       
       for (int i = 0; i < source.length; i++)
       {
           source[i] = (i < (source.length - first)) ? tmp2[i] : 
                                                       tmp1[i - (source.length - first)];
       }
       
       return source;
   }
   
   /**
    * Convert the array of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    array
    *           The array of wrapper objects.  Must not be <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static int[] objectToPrimitive(Integer[] array)
   {
      int[] out = new int[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = array[i].intValue();
      }
      
      return out;
   }
   
   /**
    * Convert the array of primitives into the corresponding array of wrapper
    * objects.
    *
    * @param    array
    *           The array of primitive types.  Must not be <code>null</code>.
    *
    * @return   The corresponding wrapper object array.
    */
   public static Integer[] primitiveToObject(int[] array)
   {
      Integer[] out = new Integer[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = Integer.valueOf(array[i]);
      }
      
      return out;
   }
   
   /**
    * Convert a collection of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    list
    *           The collection of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static int[] integerCollectionToPrimitive(Collection<Integer> list)
   {
      return objectToPrimitive(list.toArray(new Integer[list.size()]));
   }

   /**
    * Convert a map of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    map
    *           The map of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive arrays. Array on index 0 will hold
    *           the keys and array on index 1 will hold the values.
    */
   public static int[][] integerMapToPrimitive(Map<Integer, Integer> map)
   {
      int[][] res = new int[2][map.size()];
      int idx = 0;
      for (int key : map.keySet())
      {
         res[0][idx] = key;
         res[1][idx] = map.get(key);
         idx++;
      }
      return res;
   }

   /**
    * Convert an array of primitive types into a list of the corresponding
    * wrapper objects.
    *
    * @param    array
    *           The primitive array.  Must not be <code>null</code>.
    *
    * @return   The list of wrapper objects.
    */
   public static List<Integer> primitiveToIntegerList(int[] array)
   {
      return Arrays.asList(primitiveToObject(array));
   }
   
   /**
    * Check if a specific network end point can be connected from the local host.
    *
    * @param    host
    *           The given resolved host 
    * @param    port
    *           The target port number on the given host to check for availability 
    * @param    timeout
    *           The connection timeout in milliseconds 
    * 
    * @return   True if the port is listened on the remote host, otherwise false
    * 
    * @throws   SocketTimeoutException
    *           It happens if timeout expires before connecting test completed
    */
   public static boolean isEndPointAvailable(InetAddress host, int port, int timeout)
   throws SocketTimeoutException
   {
      try(Socket sock = new Socket())
      {
         sock.setReuseAddress(true);
         sock.setSoLinger(false, 0);
         sock.setTcpNoDelay(true);
         sock.connect(new InetSocketAddress(host, port), timeout);
         return true;
      }
      catch(SocketTimeoutException ex0)
      {
         throw ex0;
      }
      catch (IOException ex1)
      {
         LOG.log(Level.WARNING, "not available host:port " + host + ":" + port, ex1);
      }

      return false;
   }
   
   /**
    * Test if a specific network host name or IP address represents a network interface.
    *
    * @param    host
    *           The target network host name or IP address to check 
    * 
    * @return   True if the host represents a network interface, otherwise false
    */
   public static boolean isValidHostNameOrIpAddress(String host)
   {
      try
      {
         InetAddress addr = InetAddress.getByName(host);
         if (addr.isLoopbackAddress() || addr.isAnyLocalAddress())
         {
            return false;
         }
         
         return NetworkInterface.getByInetAddress(addr) != null;
      }
      catch(Throwable e)
      {
         return false;
      }
   }
   
   /**
    * Convert the array of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    array
    *           The array of wrapper objects.  Must not be <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static long[] objectToPrimitive(Long[] array)
   {
      long[] out = new long[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = array[i].longValue();
      }
      
      return out;
   }
   
   /**
    * Convert the array of primitives into the corresponding array of wrapper
    * objects.
    *
    * @param    array
    *           The array of primitive types.  Must not be <code>null</code>.
    *
    * @return   The corresponding wrapper object array.
    */
   public static Long[] primitiveToObject(long[] array)
   {
      Long[] out = new Long[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = Long.valueOf(array[i]);
      }
      
      return out;
   }
   
   /**
    * Convert a collection of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    list
    *           The collection of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static long[] longCollectionToPrimitive(Collection<Long> list)
   {
      return objectToPrimitive(list.toArray(new Long[list.size()]));
   }
   
   /**
    * Convert an array of primitive types into a list of the corresponding
    * wrapper objects.
    *
    * @param    array
    *           The primitive array.  Must not be <code>null</code>.
    *
    * @return   The list of wrapper objects.
    */
   public static List<Long> primitiveToLongList(long[] array)
   {
      return Arrays.asList(primitiveToObject(array));
   }
   
   /**
    * Load and return a <code>Properties</code> object from a properties
    * file specified by <code>filename</code>.
    *
    * @param   filename
    *          Name of the properties file.
    *
    * @return  Properties object containing the data read from
    *          <code>filename</code>.
    *
    * @throws  IOException
    *          if any error occurs reading the file.
    */
   public static Properties loadProperties(String filename)
   throws IOException
   {
      Properties props = new Properties();
      FileInputStream fis = null;
      BufferedInputStream bis = null;
      try
      {
         // Read properties from file.
         fis = new FileInputStream(filename);
         bis = new BufferedInputStream(fis);
         props.load(bis);
      }
      finally
      {
         // Clean up streams, regardless of success.
         if (bis != null)
         {
            bis.close();
         }
         else if (fis != null)
         {
            fis.close();
         }
      }
      
      return props;
   }


   /**
    * Displays a prompt on <code>stdout</code> and returns the input entered
    * by the user on <code>stdin</code>.  The first new line encountered
    * will end the input reading loop and the text will not include this
    * new line character.
    *
    * @param    txt
    *           The text message to display.
    *
    * @return   The text that was interactively entered or <code>null</code>
    *           if there was no input provided.
    *           
    * @throws IOException 
    */
   public static char[] prompt(String txt) 
   throws IOException 
   {
      System.out.print(txt);
      System.out.flush();
      
      StringBuilder sb = new StringBuilder();
      
      int ch;

      readLoop:
      while (true)
      {
         switch (ch = System.in.read())
         {
            case -1:
            case '\n':
               break readLoop;

            case '\f':
            case '\t':
            case '\r':
               continue readLoop;

            default:
               sb.append((char) ch);
               break;
         }
      }

      if (sb.length() == 0) 
      {
         return null;
      }

      return sb.toString().toCharArray();
   }
   
   /**
    * Determine if the given throwable or one of its causes matches one of the throwable
    * types given as criteria.
    *
    * @param    thr
    *           Throwable from which to examine cause chain.
    * @param    criteria
    *           List of throwables to check for.
    *
    * @return   <code>true</code> if any of the criteria throwables matches the given
    *           throwable or any throwable in the cause chain.
    */
   public static boolean causeChainContains(Throwable thr, Class[] criteria)
   {
      Throwable next = thr;
      do
      {
         for (int i = 0; i < criteria.length; i++)
         {
            if (criteria[i].isInstance(next))
            {
               return true;
            }
         }
         
         next = next.getCause();
      }
      while (next != null);
      
      return false;
   }
   
   /**
    * Reports the default character set name. Since the Progress default
    * charset is ISO-8859-1, when the JVM reports that UTF-8 is the default,
    * this method will return the string "ISO-8859-1" instead.
    * <p>
    * The assumption (possibly not always correct) is that UTF-8 is not a
    * valid Progress charset and instead, to provide a proper default, we
    * must override with a compatible charset. Note that if the default
    * charset name is not UTF-8, it is assumed that the charset has been
    * overridden already (at JVM startup).  In that case, the default is
    * returned unchanged.
    *
    * @return   A default charset name.
    */
   public static String getCharsetName()
   {
      return getCharsetOverride();
   }
   
   /**
    * Overrides the default character set by the Progress default charset ISO-8859-1 
    * if UTF-8 is the default JVM charset, otherwise returns the JVM default charset.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   The JVM default charset if no override is required.
    */
   public static String getCharsetOverride()
   {
      Charset def  = Charset.defaultCharset();
      return getCharsetOverride(def.name());
   }

   /**
    * Overrides the given UTF-8 charset by the Progress default charset ISO-8859-1, otherwise
    * returns the given charset unchanged.
    * 
    * @param    charsetName
    *           Charset name to override
    *
    * @return   The given charset name  if no override is required.
    */
   public static String getCharsetOverride(String charsetName)
   {
      String  cset;
      
      // this test may not be complete, but it handles the most common
      // case where the extended ASCII range will be "munged" by an
      // improper byte to char conversion
      if ("UTF-8".equalsIgnoreCase(charsetName))
      {
         // this gets called too much to allow it to log since we don't want it cluttering
         // client or conversion logs
         // token = UnimplementedFeature.todo("This code doesn't properly handle UTF-8.", token);
         
         // override with a safe default (matches the Progress default) 
         cset = "ISO-8859-1";
      }
      else
      {
         cset = charsetName;
      }
      
      return cset;
   }
   

   /**
    * Convert the given byte to a Unicode character using the given character
    * set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * The normal ASCII range can be converted to a Unicode character by a
    * simple widening conversion from byte to char. But the extended ASCII
    * range (128 through 255 inclusive) will translate to character values
    * that cannot be achieved via a widening conversion.  Instead the
    * proper character set must be used to get the right result.
    *
    * @param    b
    *           The byte to convert.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The Unicode character that represents the given byte in the
    *           specified character set. In case of any error, -1 will be
    *           returned.
    */
   public static char toChar(byte b, String cset)
   {
      String txt = null;
      char   ch  = (char) -1;
      
      try
      {
         if (cset == null)
         {
            // use the default character set to translate this byte
            // into a character
            txt = new String(new byte[] {b});
         }
         else
         {
            // use a specific character set to convert rather than the
            // default
            txt = new String(new byte[] {b}, cset);
         }
         
         ch = txt.charAt(0);
      }
      
      catch (UnsupportedEncodingException uee)
      {
         // nothing to do
      }
      
      return ch;
   }
   
   /**
    * Convert the given byte to a Unicode character using the given character
    * set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * <b>Only the least significant byte will be converted, all other data is
    * dropped at this time.</b>
    * <p>
    * The normal ASCII range can be converted to a Unicode character by a
    * simple widening conversion from byte to char. But the extended ASCII
    * range (128 through 255 inclusive) will translate to character values
    * that cannot be achieved via a widening conversion.  Instead the
    * proper character set must be used to get the right result.
    *
    * @param    keyCode
    *           The number to convert. Must be between 0 and 255 inclusive.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The Unicode character that represents the given byte in the
    *           specified character set. In case of any error, -1 will be
    *           returned.
    */
   public static char toChar(int keyCode, String cset)
   {
      try
      {
         int codepoint = getCodepoint(keyCode, cset);
         
         if (codepoint != -1)
         {
            char[] chars = Character.toChars(codepoint);
            
            if (chars.length == 1)
            {
               return chars[0];
            }
         }
      }
      catch (IllegalArgumentException ex)
      {
         // no-op
      }
      
      return (char) -1;
   }
   
   /**
    * Convert the given Unicode character to an integer value using the given
    * character set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * The normal ASCII range can be converted from a Unicode character by a
    * simple narrowing conversion from char to byte. But the Unicode
    * characters representing the extended ASCII range (128 through 255
    * inclusive) are not represented in Unicode via a single byte. So a
    * translation from character values to a number cannot be achieved via a
    * narrowing conversion.  Instead the proper character set must be used to
    * get the right result.
    *
    * @param    ch
    *           The Unicode character to convert.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The integer value that represents the given the Unicode 
    *           character in the specified character set or -1 on any error.
    */
   public static int toInt(char ch, String cset)
   {
      int i = -1;
      
      try
      {
         String encoded = new String(new char[] {ch});
         byte[] b = (cset == null) ? encoded.getBytes() : encoded.getBytes(cset);
         i = b[0] & 0x000000FF;
      }
      
      catch (UnsupportedEncodingException uee)
      {
         // let -1 be returned
      }
      
      return i;
   }
   
   /**
    * Convert a unicode codepoint into codepoint according to the provided charset.
    * 
    * @param    codepoint
    *           The given unicode codepoint
    * @param    cset
    *           The target charset name
    * 
    * @return   The integer code representing the given unicode codepoint in the target charset.
    * 
    * @throws   CharacterCodingException
    *           If the given unicode codepoint cannot be represented by the target charset
    * @throws   UnsupportedCharsetException
    *           If the target charset is not supported. 
    */
   public static int convertCodepoint(int codepoint, String cset)
   throws CharacterCodingException, UnsupportedCharsetException
   {
      char[] keyChars = Character.toChars(codepoint);
      Charset charSet = Charset.forName(cset);
      CharsetEncoder encoder = charSet.newEncoder();
      encoder.onMalformedInput(CodingErrorAction.REPORT);
      encoder.onUnmappableCharacter(CodingErrorAction.REPORT);
      ByteBuffer byteBuffer = encoder.encode(CharBuffer.wrap(keyChars));
      byte[] keyBytes = byteBuffer.array();
      int len = keyBytes.length;
      
      for(int i = len - 1; i >= 0; i--)
      {
         if (keyBytes[i] != 0)
         {
            len = i;
            break;
         }
      }
      
      int ret = 0;
      
      for (int i = 0; i <= len; i++)
      {
         int bt = Byte.toUnsignedInt(keyBytes[i]);
         
         //System.out.println(Integer.toString(bt, 16));
         
         ret = ret | bt;
         
         if (i < len)
         {
            ret = ret << 8;
         }
      }
      
      return ret;
   }

   /**
    * Get the unicode codepoint for the given key code and char set.
    * 
    * @param    keyCode
    *           The given key code
    * @param    cset
    *           The given char set.
    * 
    * @return   The unicode codepoint if keycode is a valid key code in the given char set, otherwise -1. 
    */
   public static int getCodepoint(int keyCode, String cset)
   {
      ByteBuffer bf = ByteBuffer.allocate(4);
      bf.putInt(keyCode);
      bf.flip();
      try
      {
         for (int i = 0; i < 4; i++)
         {
            if (bf.get(i) != 0)
            {
               bf.position(i);
               break;
            }
         }
         
         byte[] buf = new byte[bf.limit() - bf.position()];
         
         for (int i = bf.position(), j = 0; i < bf.limit(); i++, j++)
         {
            buf[j] = bf.get(i);
         }
         
         Charset charSet = Charset.forName(cset);
         CharsetDecoder decoder = charSet.newDecoder();
         decoder.onMalformedInput(CodingErrorAction.REPORT);
         decoder.onUnmappableCharacter(CodingErrorAction.REPORT);
         CharBuffer charBuffer = decoder.decode(ByteBuffer.wrap(buf));
         
         String key = new String(charBuffer.array());
         
         return Character.codePointAt(key, 0);
      }
      catch (IllegalArgumentException | CharacterCodingException e)
      {
      }
      
      return -1;
   }

   /**
    * Creates all subdirectories needed to provide a valid path for the
    * target filename that is passed in.  It is expected that there will 
    * always be a path included in this filename such that making 
    * directories makes sense.
    *
    * @param    tar
    *           Target filename for which paths must be created.
    *
    * @return   <code>true</code> if the full necessary path already
    *           exists or if it has been created, <code>false</code>
    *           otherwise.
    */
   public static boolean createPathForTarget(String tar)
   {
      String path = tar.substring(0, tar.lastIndexOf(File.separatorChar));
      File   file = new File(path);
      
      // anything to do?
      if (file.exists() && file.isDirectory())
         return true;
      
      try
      {
         // best effort to create the directories including all needed parent dirs
         file.mkdirs();
      }
      catch (Exception exc)
      {
         // TODO: add logging
         // some failure, probably a security violation
         return false;
      }
      
      // it better be there now...
      if (file.exists() && file.isDirectory())
         return true;

      // some other unexpected problem!
      return false;
   }
   
   /**
    * Find the specified file in the file system using the specified search
    * paths and return it. Handles absolute and relative paths.
    *
    * @param   filename
    *          Name of the source file to find; if this is not an absolute
    *          path, it should be relative to one of the paths in the
    *          <code>searchPaths</code> array.
    * @param   paths
    *          Array of absolute paths which are prepended to a relative
    *          <code>filename</code>.
    *
    * @return  The file if it was found.
    *
    * @throws  FileNotFoundException
    *          if a file named <code>filename</code> does not exist in the
    *          specified search paths (or at the absolute path provided).
    */
   public static File findFileInPath(String filename, String[] paths)
   throws FileNotFoundException
   {
      File file = new File(filename);
      
      if (file.isAbsolute())
      {
         // Return file if it is absolute and exists.
         if (file.exists())
         {
            return file;
         }
      }
      else
      {
         // Use search paths to resolve relative references.
         int len = paths.length;
         for (int i = 0; i < len; i++)
         {
            file = new File(paths[i], filename);
            if (file.exists())
            {
               return file;
            }
         }
      }
      
      // Reassemble path for error message.
      StringBuilder buf = new StringBuilder();
      String sep = File.pathSeparator;
      int len = paths.length;
      for (int i = 0; i < len; i++)
      {
         buf.append(paths[i])
            .append(sep);
      }
      
      throw new FileNotFoundException("File '" + filename + "' not found in path: " + buf);
   }
   
   /**
    * Calculates a resource path to the target page within its application class path.
    * 
    * @param    pageContext
    *           The page root path
    * @param    location
    *           The package where the target resource is located. 
    * @param    pageFile
    *           The resource name with its extension.
    * 
    * @return   The string path to be found within its application class path.
    */
   public static String calculateResourcePath(String   pageContext,
                                              Package  location,
                                              String   pageFile)
   {
      StringBuilder sb = new StringBuilder(pageContext);
      
      String packageName = location.getName();
      
      sb.append(packageName.replace('.', '/'));
      
      if (!pageFile.startsWith("/"))
      {
         sb.append("/");
      }
      
      return sb.append(pageFile).toString();
   }
   
   /**
    * Canonicalize the given file system path as described by
    * <code>java.io.File.getCanonicalPath()</code>.  In addition, normalize
    * the use of the file separator by replacing all instances of
    * <code>oldSeparator</code> with <code>newSeparator</code>.
    * 
    * @param   path
    *          A relative or absolute path to be canonicalized.
    * @param   oldSeparator
    *          File separator to be replaced.
    * @param   newSeparator
    *          New file separator.
    *
    * @return  Canonicalized path containing normalized file separators.
    * 
    * @throws  IOException
    *          If there is a file system while canonicalizing the path.
    */
   public static String canonicalizePath(String path, String oldSeparator, String newSeparator)
   throws IOException
   {
      String res = null;
      try
      {
         res = new File(path).getCanonicalPath();
      }
      catch (IOException e)
      {
         // On windows trying to access some path could throw exceptions like these:
         // TODO: identify correct fix and remove this line
         LOG.warning("", e);
         
         res = path; // consider path as already canonical (or return an invalid value ?)
      }
      
      res = replaceSeparator(res, oldSeparator, newSeparator);
      
      return res;
   }
   
   /**
    * Replace all instances of {@code oldSeparator} with {@code newSeparator} in the given path.
    * 
    * @param   path
    *          A file path.
    * @param   oldSeparator
    *          File separator to be replaced.
    * @param   newSeparator
    *          New file separator.
    *
    * @return  Path containing normalized file separators.
    */
   public static String replaceSeparator(String path, String oldSeparator, String newSeparator)
   {
      if (!oldSeparator.equals(newSeparator))
      {
         if ("\\".equals(oldSeparator))
         {
            oldSeparator = "\\\\";
         }
         if ("\\".equals(newSeparator))
         {
            newSeparator = "\\\\";
         }
         path = path.replaceAll(oldSeparator, newSeparator);
      }
      
      return path;
   }
   
   /**
    * Determines if a both source and target exist as files and if the target
    * has a later timestamp.  It does not consider contents or size in this
    * determination.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    *
    * @return   <code>true</code> if the both files exist and the target is
    *           newer, otherwise <code>false</code> (includes failures).
    */
   public static boolean isLaterVersion(String source, String target)
   {      
      File src = new File(source);
      File tar = new File(target);
      
      // make sure the source file is OK
      boolean srcOK = src.exists() && src.isFile();
      
      if (!srcOK)
         return false;
      
      // make sure the target file is not there or if it is there that
      // it is older than the source
      if (tar.exists())
      {
         if (tar.isFile())
         {
            // returns true if the source is older, false if the target is
            // older
            return src.lastModified() < tar.lastModified();
         }
      }
      
      return false;
   }     
   
   /**
    * Creates the directory tree to the target if needed and then copies
    * the source to the target if needed.  The copy is byte for byte, no
    * modifications of the file will occur.
    *
    * @param    src
    *           Source filename.
    * @param    tar
    *           Target filename.
    *
    * @return   <code>true</code> if the file was copied successfully OR
    *           if the file didn't need to be copied, <code>false</code>
    *           on any failure.
    */
   public static boolean simpleCopy(String src, String tar)
   {
      // try to create the directory
      if (!createPathForTarget(tar))
      {
         // dir creation failed
         return false;
      }
      
      return copyFileIfNeeded(src, tar, 0);
   }
   
   /**
    * Conditionally copies a file, byte for byte (no changes) to a new file,
    * if and only if the source is an accessible file and the target doesn't
    * exist or the target exists as a file but is older than the source or
    * has a different size.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    * @param    bsize
    *           Buffer size or if 0 the buffer size will default to the file
    *           size or 64Kb whichever is less.
    *
    * @return   <code>true</code> if the file was copied successfully OR if
    *           the file didn't need to be copied, <code>false</code> on any
    *           failure.
    */
   public static boolean copyFileIfNeeded(String source, String target, int bsize)
   {      
      File src = new File(source);
      File tar = new File(target);
      
      // make sure the source file is OK
      boolean srcOK = src.exists() && src.isFile();
      
      if (!srcOK)
         return false;
      
      boolean tarOK = true; 
      
      // make sure the target file is not there or if it is there that
      // it is older than the source or it is not the same size, before
      // trying to copy
      if (tar.exists())
      {
         if (tar.isFile())
         {
            boolean oldSrc = src.lastModified() > tar.lastModified();
            long    diff   = src.length() - tar.length();               
            tarOK = oldSrc || (diff != 0);
         }
         else
         {
            return false;
         }
      }
      
      boolean result = false;
      
      // already checked src, now check target
      if (tarOK)
      {
         // best efforts attempt to copy
         result = copyFile(source, target, 0, false);
      }
      
      return result;
   }   
   
   /**
    * Handles the byte for byte copying of a single source file to a target 
    * file, using a buffer size of the caller's choice. The resulting target
    * file will be a newly created file (date/timestamps will be new).
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    * @param    bsize
    *           Buffer size or if 0 the buffer size will default to the file
    *           size or 64Kb whichever is less.
    * @param    append
    *           <code>true</code> if the output file should be opened in
                append mode.
    *
    * @return   <code>true</code> if the file was copied successfully,
    *           <code>false</code> on any failure.
    */
   public static boolean copyFile(String  source, String target, int bsize, boolean append)
   {
      File in  = new File(source);
      long len = in.length();
      
      // if an invalid buffer size is set, the lesser of 64Kb or the file
      // length is used
      if (bsize < 1)
      {
         bsize = 64*1024;
         
         if (len < bsize)
            bsize = (int) len;
      }
      
      try
      {
         FileInputStream  src = new FileInputStream(in);
         FileOutputStream tar = new FileOutputStream(target, append);
         
         int off = 0;
         
         byte[] buf = new byte[bsize];
   
         while (len > bsize)
         {
            src.read(buf, off, bsize);
            tar.write(buf, 0, bsize);
            len -= bsize;
         }
         
         src.read(buf, off, (int)len);
         tar.write(buf, 0, (int)len);
         
         src.close();
         tar.close();
      }
      catch (Exception exc)
      {
         return false;
      }
      
      return true;
   }
   
   /**
    * Supply a list of files which case-insensitively match the given file
    * name.  On a case-insensitive file system, this list will contain no more
    * than one match.  However, on a case-sensitive file system, the list may
    * contain more matches.
    * 
    * @param   path
    *          The file name to be matched.  It will be canonicalized before
    *          the search is conducted.
    * @param   separator
    *          Legacy system file separator which will be replaced with the
    *          target system's file separator (in the result list).
    *          
    * @return  A list containing 0 or more files.  This method will never
    *          return <code>null</code>.
    *          
    * @throws  IOException
    *          if there is any error canonicalizing the <code>path</code>.
    */
   public static List<File> getCaseInsensitiveFilenameMatches(String path, String separator)
   throws IOException
   {
      // canonicalize path and ensure file separators are appropriate for the
      // current platform's file system
      String canonicalPath = canonicalizePath(path, separator, File.separator);
      
      return getCaseInsensitiveFilenameMatches(canonicalPath);
   }
   
   /**
    * Supply a list of files which case-insensitively match the given file
    * name.  On a case-insensitive file system, this list will contain no more
    * than one match.  However, on a case-sensitive file system, the list may
    * contain more matches.
    * 
    * @param   path
    *          The file name to be matched.  It must already be canonicalized
    *          (see {@link #canonicalizePath}) and with all file separators
    *          converted to the current system's file separator character.
    *          
    * @return  A list containing 0 or more files.  This method will never
    *          return <code>null</code>.
    *          
    * @throws  IOException
    *          if there is any error canonicalizing the <code>path</code>.
    */
   public static List<File> getCaseInsensitiveFilenameMatches(String path)
   throws IOException
   {
      List<File> list = new ArrayList<>(1);
      matchPathCaseInsensitively(null, path, list);
      
      return list;
   }


   /**
    * Get the short name of a class or interface (everything after the last
    * dot in the fully qualified class name).
    *
    * @param   cls
    *          Class.
    *
    * @return  Unqualified class name.
    */
   public static String getUnqualifiedName(Class<?> cls)
   {
      String name = cls.getName();
      
      return name.substring(name.lastIndexOf(".") + 1);
   }

   /**
    * Apply function to the list elements while the result is <code>null</code>
    * 
    * @param list
    *        the list to be traversed
    * @param fn
    *        Function to be applied
    *        
    * @return the first non-null result or <code>null</code> if none found
    * 
    */
   public static <T,S> T tryWhileNull(List<S> list, Function<S,T> fn)
   {
      return  list.stream().map(fn).
            filter(r -> r != null).findFirst().orElse(null);
   }

   /**
    * Find as many files as case-insensitively match the file specified by
    * <code>fullPath</code>, where the directory specified by
    * <code>leading</code> represents a leading portion of the given path.
    * <p>
    * This method calls itself recursively with progressively more specific
    * <code>leading</code> directories, until <code>leading</code> represents
    * the full path qualifier for the file which is the target of the match.
    * <p>
    * All paths in this method are processed using the current system's file
    * separator character.  This means that any input to this method must
    * have already been properly converted to the system file separator.
    * 
    * @param   leading
    *          A leading portion of the directory qualifier for the
    *          <code>fullPath</code> argument. The top-level call to this
    *          method would pass <code>null</code> here.  Recursive calls will
    *          pass a value. If given, this MUST be a valid directory in the
    *          file system AND it must be a valid prefix for the target file
    *          (it must case-insensitively match the begining of the target 
    *          filename).
    * @param   fullPath
    *          A canonicalized path name which represents an absolute path of
    *          the file to be matched.  Must NOT be <code>null</code>.
    * @param   list
    *          A list in which to store matching <code>File</code> objects
    *          (which is the result of this processing).  Must NOT be
    *          <code>null</code>.
    */
   private static void matchPathCaseInsensitively(File leading,
                                                  String fullPath,
                                                  List<File> list)
   {
      final String SEP = File.separator;
      int pos = 0;
      
      // process leading portion (pull out the directory name and calculate
      // its length when ending with the file separator character)
      if (leading != null)
      {
         String dirName = leading.getAbsolutePath();
         pos = dirName.length();
         
         // safety check
         if (!dirName.equalsIgnoreCase(fullPath.substring(0, pos)))
         {
            // TODO: add error reporting/logging
            
            // leading portion cannot match the target file name
            return;
         }
         
         if (!dirName.endsWith(SEP))
         {
            pos += SEP.length(); // always 1, from File.separator javadoc
         }
      }
      
      // gather everything after the leading portion
      String remaining = fullPath.substring(pos);
      
      // find the end of the next path segment; this is the last segment (which
      // is the base file name itself) if this is -1
      pos = remaining.indexOf(SEP);
      
      // extract the portion of the path which corresponds with the directory
      // level we're currently scanning; this is a path segment only
      final String match = (pos >= 0
                            ? remaining.substring(0, pos)
                            : remaining);
      
      // file filter which matches case-insensitively on the match string
      final FileFilter ff = new FileFilter()
      {
         @Override
         public boolean accept(File test)
         {
            // this only checks one segment at a time, not portions that
            // "cross" or include file separator characters
            String name = test.getName();
            if (name.isEmpty())
            {
               // special care on windows: when comparing the drive (C:\) segments the name
               // is empty because there is no directory.
               // manually comparing the drives:
               return test.getAbsolutePath().equalsIgnoreCase(match + SEP);
            }
            
            return name.equalsIgnoreCase(match);
         }
      };
      
      File[] files = null;
      
      if (leading == null)
      {
         // top-level call to this method
         
         // special case:  no leading directory specified, so we start at the
         // file system roots and try to match the portion of our path up to
         // the first file separator
         File[] roots = File.listRoots();
         List<File> l = new ArrayList<>(roots.length);
         for (File next : roots)
         {
            if (ff.accept(next))
            {
               l.add(next);
            }
         }
         
         files = l.toArray(new File[l.size()]);
      }
      else
      {
         // recursive call to this method
         
         // safety code; leading must always be a directory
         if (!leading.isDirectory())
         {
            // TODO: add error reporting/logging
            
            // this is definitely not a match
            return;
         }
         
         files = leading.listFiles(ff);
         
         if (files == null)
         {
            // TODO: add error reporting/logging
            
            // this should only occur on an I/O error
            return;
         }
      }
      
      if (pos < 0)
      {
         // last level of search; the contents of the files array contains our final result(s)
          Collections.addAll(list, files);
      }
      else
      {
         // search recursively down through next set of matching directories
         for (File next : files)
         {
            // file matches cannot be checked further using recursion and they
            // can't match anyway, so avoid them
            if (next.isDirectory())
            {
                matchPathCaseInsensitively(next, fullPath, list);
            }
         }
      }
   }
   
   /**
    * Generic type safe equals. It can be used for <code>null</code> objects
    * without fear of NPE. If both provided objects are <code>null</code> then
    * they are assumed equal.
    * @param   <T>
    *          Type parameter.
    * @param   obj1
    *          First object for comparison.
    * @param   obj2
    *          Second object for comparison.
    *
    * @return  <code>true</code> if objects are equal.
    */
   public static<T> boolean equals(T obj1, T obj2)
   {
      if (obj1 == null && obj2 == null)
      {
         return true;
      }
      
      return (obj1 != null) ? obj1.equals(obj2) : obj2.equals(obj1);
   }


   
   /**
    * Check if the two sets are the same. Note that <code>null</code> instances are allowed for 
    * either set.
    * 
    * @param    a
    *           The first set.
    * @param    b
    *           The second set.
    * 
    * @return   <code>true</code> if the sets match.
    */
   public static<T> boolean isSame(Set<T> a, Set<T> b)
   {
      if (a == null)
      {
         return b == null;
      }
      else if (b == null)
      {
         return false;
      } 
      return a.size() == b.size() && a.containsAll(b);
   }
   
   /**
    * Check if a given value is a valid value.
    * 
    * @param    value
    *           Value to check.
    * @param    validValues
    *           List of valid values.
    *          
    * @return   <code>true</code> if value is in the list of valid values,
    *           otherwise <code>false</code>.
    */
   public static boolean isValidValue(int value, int ... validValues)
   {
      for (int v : validValues)
      {
         if (value == v)
         {
            return true;
         }
      }
      
      return false;
   }

   /**
    * Determines whether two sets have common elements.
    * 
    * @param <T> 
    *
    * @param    a
    *           a set
    * @param    b
    *           another set
    *
    * @return   <code>true</code> if there are common elements
    */
   public static<T> boolean intersect(Set<T> a, Set<T> b)
   {
      if (a == null || b == null || a.isEmpty() || b.isEmpty())
      {
         return false;
      }
      
      int aSz  = a.size();
      int bSz  = b.size();
      
      // iterate the smaller of the two sets
      Iterator<T> iter = aSz <= bSz ? a.iterator() : b.iterator();
      Set<T>      comp = aSz <= bSz ? b : a;
         
      while (iter.hasNext())
      {     
         if (comp.contains(iter.next()))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Produces both the intersection and difference of two sets and puts those values into the 
    * given result set instances.
    * 
    * @param    <T>
    *           The type of the set elements.
    * @param    a
    *           a set
    * @param    b
    *           another set
    * @param    inter
    *           The result set which will contain the intersection results.
    * @param    diff
    *           The result set which will contain the difference results.  This is the set of items in "a"
    *           which are not in "b", but it does not include items in "b" which are not in "a".
    */
   public static<T> void split(Set<T> a, Set<T> b, Set<T> inter, Set<T> diff)
   {
      inter.clear();
      diff.clear();
      
      Iterator<T> iter = a.iterator();
      T o = null;
      
      while (iter.hasNext())
      {
         o = iter.next();
         
         if (b.contains(o))
         {
            inter.add(o);
         }
         else
         {
            diff.add(o);
         }
      }
   }   
   
   /**
    * Produces the intersection of two sets and puts it into the third set.
    * 
    * @param <T> 
    *
    * @param    a
    *           a set
    * @param    b
    *           another set
    * @param    result
    *           result set
    */
   public static<T> void intersection(Set<T> a, Set<T> b, Set<T> result)
   {
      result.clear();
      
      int aSz  = a.size();
      int bSz  = b.size();
      
      // iterate the smaller of the two sets
      Iterator<T> iter = aSz <= bSz ? a.iterator() : b.iterator();
      Set<T>      comp = aSz <= bSz ? b : a;
      
      while (iter.hasNext())
      {
         T t = iter.next();
         
         if (comp.contains(t))
         {
            result.add(t);
         }
      }
   }
   
   /**
    * Produces the difference of two sets and puts it into the third set.
    *
    * @param    <T> 
    *           The element type.
    * @param    a
    *           The first set.
    * @param    b
    *           The set to subtract from 'a'.
    * @param    result
    *           The result set which will contain the difference results.  This is the set of items in "a"
    *           which are not in "b", but it does not include items in "b" which are not in "a".
    */
   public static<T> void difference(Set<T> a, Set<T> b, Set<T> result)
   {
      result.clear();
      
      Iterator<T> iter = a.iterator();
      T o = null;
      
      while (iter.hasNext())
      {
         o = iter.next();
         
         if (!b.contains(o))
         {
            result.add(o);
         }
      }
   }
   
   /**
    * Gets a resource path string to be cleared from invalid sequences and file separator duplication.
    * Handles properly current('.') and uplevel('..') directory entries and empty spaces.  No case
    * letter changes are performing.
    * 
    * @param    nameDirty
    *           The not yet refined resource path string.
    * @param    separator
    *           The string representing file separator to use in result.
    *
    * @return   Properly formatted case sensitive image path string.
    */
   public static String getRefinedPathName(String nameDirty, String separator)
   {
      // parse the dirty path into array by separator to clean for multiple separators
      String[] splitPath = nameDirty.split(separator);
      int pathSize = splitPath.length;
      
      // walk through the entries to clean up
      for (int i = 0; i < pathSize; i++)
      {
         // process next path name entry
         String name = splitPath[i];
         
         // skip the current directory, empty directory, directory or files with more than 2 dots
         // in beginning
         if (name.equals(".") || name.isEmpty() || (name.startsWith("..") && !name.equals("..")))
         {
            // go one directory level up
            for (int j = i; j < pathSize - 1; j++)
            {
               splitPath[j] = splitPath[j + 1];
            }
            // one entry has been removed for current dir
            pathSize--;
            i--;
            
            continue;
         }
         
         // try to handle parent directory shift
         if (name.equals(".."))
         {
            // for the first and last entries just ignore this - no parent directory
            if (i > 0 && i < pathSize - 1)
            {
               // go one directory level up
               for (int j = i; j < pathSize - 1; j++)
               {
                  splitPath[j - 1] = splitPath[j + 1];
               }
               i--;
            }
            // two entries have been removed, one for '..' and parent directory
            pathSize -= 2;
            continue;
         }
         
      }
      
      // second pass - build the final buffer
      StringBuilder sb = new StringBuilder();
      // at this time we must have clean array
      for (int i = 0; i < pathSize; i++)
      {
         sb.append(splitPath[i]);
         
         // for all entries except last add the separator
         if (i < pathSize - 1)
         {
            sb.append(separator);
         }
      }
      
      // fix the directory path 
      if (nameDirty.endsWith(separator) && !sb.toString().endsWith(separator))
      {
         sb.append(separator);
      }
      
      return sb.toString();
   }

   /**
    * Get the path of the jar file containing the given class.
    * 
    * @param    clazz
    *           The given class
    * 
    * @return   The path of the jar file if it exists, otherwise null.
    */
   public static String getJarPath(Class clazz)
   {
      CodeSource codeSource = clazz.getProtectionDomain().getCodeSource();
      File jarFile;
      try
      {
         if (codeSource.getLocation() != null)
         {
            jarFile = new File(codeSource.getLocation().toURI());
         }
         else
         {
           String path = clazz.getResource(clazz.getSimpleName() + ".class").getPath();
           String jarFilePath = path.substring(path.indexOf(":") + 1, path.indexOf("!"));
           jarFilePath = URLDecoder.decode(jarFilePath, "UTF-8");
           jarFile = new File(jarFilePath);
         }
      }
      catch (URISyntaxException | UnsupportedEncodingException e)
      {
         LOG.severe("", e);
         return null;
      }
      
      return jarFile.getPath();
   }
   
   /**
    * Produces the difference of two int arrays.
    *
    * @param    a
    *           The first array.
    * @param    b
    *           The second array which is to be subtracted from the first.
    *
    * @return   All elements from the first array that are not in the
    *           second array.
    */
   public static int[] diffIntArrays(int[] a, int[] b)
   {
      // boundary cases
      if (a.length == 0)
      {
         return new int[] {};
      }

      int[] aa = new int[a.length];
      for (int i = 0; i < a.length; i ++)
      {
         aa[i] = a[i];
      }
         
      Arrays.sort(aa);

      if (b.length == 0)
      {
         return aa;
      }
      
      int[] bb = new int[b.length];
      for (int i = 0; i < b.length; i ++)
      {
         bb[i] = b[i];
      }
         
      Arrays.sort(bb);

      int[] res = new int[aa.length];
      int count = 0;
      int i = 0;
      int j = 0;
      
      scan:
      for (i = 0; i < aa.length; i ++)
      {
         // equal numbers; skip both
         if (aa[i] == bb[j])
         {
            j ++;
            if (j < bb.length)
            {
               continue;
            }
            
            break scan;
         }
         
         // integer from a to keep
         if (a[i] < b[j])
         {
            res[count ++] = a[i];
            continue;
         }
         
         // integer from b to skip
         while (a[i] > b[j])
         {
            j ++;
            if (j < bb.length)
            {
               continue;
            }
            
            break scan;
         }
      }

      // copy the rest of a
      for (int k = i + 1; k < aa.length; k ++)
      {
         res[count ++] = aa[k];
      }
            
      // compact the array
      int[] diff = new int[count];
      for (i = 0; i < count; i ++)
      {
         diff[i] = res[i];
      }
      
      return diff;
   }
   
   /**
    * The method returns a reverse {@linkplain Stream}
    * from the supplied <code>array</code>.
    * 
    * @param    array
    *           The array to be reversed in a stream.
    *           
    * @return   see above.
    */
   public static <T> Stream<T> reverseStream(T[] array) 
   {
      int size = array.length;
      return IntStream.rangeClosed(0, size - 1)
                      .mapToObj(i -> array[size - i - 1]);
   }
   
   /**
    * The method returns a reverse {@linkplain Stream}
    * from the supplied linked list.
    * 
    * @param    list
    *           The linked list to be reversed in a stream.
    *           
    * @return   see above.
    */
   public static <T> Stream<T> reverseStream(LinkedList<T> list) 
   {
      Iterator<T> it = list.descendingIterator();
      Spliterator<T> sit = Spliterators.spliteratorUnknownSize(it, Spliterator.ORDERED);
      return StreamSupport.stream(sit, false);
   }
   
   /**
    * The method returns a reverse {@linkplain Stream}
    * from the supplied <code>list</code>.
    * 
    * @param    list
    *           The list to be reversed in a stream.
    *           
    * @return   see above.
    */
   public static <T> Stream<T> reverseStream(List<T> list) 
   {
      if (list instanceof LinkedList<?>) 
      {
         return reverseStream((LinkedList<T>) list);
      }
      else
      {
         int size = list.size();
         return IntStream.rangeClosed(0, size - 1)
                         .mapToObj(i -> list.get(size - i - 1));
      }
   }
   
   /**
    * The method returns the first non-null argument from the supplied array of 
    * arguments.
    * 
    * @param    args
    *           An array of arguments.
    *           
    * @return   See above.
    * 
    * @throws   IllegalArgumentException
    *           when zero arguments is passed to the method.
    * @throws   NullPointerException
    *           when no non-null value is found in the supplied array of arguments.
    */
   @SafeVarargs
   public static <T> T nonNull(T... args)
   {
      if (args == null || args.length == 0)
      {
         throw new IllegalArgumentException();
      }
      
      for (T item : args)
      {
         if (item != null)
         {
            return item;
         }
      }
      
      throw new NullPointerException();
   }
   
   /**
    * The method returns the first non-empty trimmed string represented by the supplied
    * arguments. {@linkplain Object#toString()} is used to convert each argument to a 
    * string value and trimmed by {@linkplain String#trim()}.
    * 
    * @param    args
    *           An array of arguments.
    *           
    * @return   See above.
    * 
    * @throws   IllegalArgumentException
    *           when zero arguments is passed to the method or no non-empty string is 
    *           represented by any of the supplied arguments.
    */
   @SafeVarargs
   public static String nonEmpty(Object... args)
   {
      if (args == null || args.length == 0)
      {
         throw new IllegalArgumentException();
      }
      
      for (Object item : args)
      {
         if (item == null)
         {
            continue;
         }
         
         String str = item.toString().trim();
         if (!str.isEmpty())
         {
            return str;
         }
      }
      
      throw new IllegalArgumentException();
   }

   /**
    * Converts string size value to integer one.  The string can be suffixed with MB or KB or none
    * for megabyes base init.  The input string must be non null and not empty.
    *           
    * @param    sizeStr
    *           An string representing the integer limit value.
    * 
    * @return   see above.
    */
   public static int getSizeLimitFromString(String sizeStr)
   {
      // default unit will be bytes
      int res = 1;
      
      // prepare string
      sizeStr = sizeStr.toUpperCase().trim();

      // analyze the string
      StringBuilder sb = new StringBuilder();
      for (int i = 0; i < sizeStr.length(); i++)
      {
         char chNext = sizeStr.charAt(i);
         if (Character.isDigit(chNext))
         {
            sb.append(chNext);
         }
         else
         {
            if (chNext == 'M')
            {
               // the base units are megabytes
               res = 1024 * 1024;
               break;
            }
            else if (chNext == 'K')
            {
               // the base units are kilobytes
               res = 1024;
               break;
            }
            else
            {
               // ignore other chars
               continue;
            }
         }
      }

      // now string builder must have only numbers
      res *= Integer.valueOf(sb.toString()).intValue();
      
      return res;
   }

   /**
    * Converts long number to little-endian bytes.
    *
    * @param   l
    *          A long number.
    *
    * @return  the returned byte array.
    */
   public static byte[] longToBytesLE(long l)
   {
      return new byte[]
      {
         (byte) l,
         (byte) (l >> 8),
         (byte) (l >> 16),
         (byte) (l >> 24),
         (byte) (l >> 32),
         (byte) (l >> 40),
         (byte) (l >> 48),
         (byte) (l >> 56)
      };
   }

   /**
    * Converts little-endian bytes to long number.
    *
    * @param   data
    *          Bytes to convert.
    *
    * @return  the returned long.
    */
   public static long bytesToLongLE(byte[] data)
   {
      long value = data[0] & 0xffL;
      value += (data[1] & 0xffL) << 8;
      value += (data[2] & 0xffL) << 16;
      value += (data[3] & 0xffL) << 24;
      value += (data[4] & 0xffL) << 32;
      value += (data[5] & 0xffL) << 40;
      value += (data[6] & 0xffL) << 48;
      value += (data[7] & 0xffL) << 56;
      return value;
   }


   /**
    * Collect all the implemented interfaces (directly and from the hierarchy), for the given
    * class.
    * <p>
    * The collection is performed doing a BFS walk of the hierarchy, starting from the given class.
    * 
    * @param    cls
    *           The class from where to get the interfaces.
    * @param    interfaces
    *           A set where to store the interfaces, in the BFS order..
    */
   public static void collectInterfaces(Class<?> cls, Set<Class<?>> interfaces)
   {
      collectInterfacesImpl(cls, interfaces, new HashSet<>());
   }
   
   /**
    * Do a DFS walk of the class hierarchy and execute the given task when going 'back up' on the tree.
    * 
    * @param    cls
    *           The class to start the walk.
    * @param    task
    *           The task to execute.
    */
   public static void dfsClassHierarchy(Class<?> cls, Consumer<Class<?>> task)
   {
      Class<?>[] ifcs = cls.getInterfaces();
      if (ifcs != null)
      {
         for (Class<?> ifc : ifcs)
         {
            dfsClassHierarchy(ifc, task);
         }
      }

      if (cls.getSuperclass() != null)
      {
         dfsClassHierarchy(cls.getSuperclass(), task);
      }
      
      task.accept(cls);
   }

   /**
    * Reads a row of CSV values from the reader and returns the values as a Java list.
    *
    * @param   r
    *          Reader to read from.
    *
    * @return  A valid list or {@code null} if end of stream is reached.
    * 
    * @throws IOException 
    */
   public static List<String> readCSVLine(Reader r)
   throws IOException
   {
      int ch = r.read();

      // skip CR
      while (ch == '\r')
      {
         ch = r.read();
      }

      if (ch < 0)
      {
         return null;
      }

      List<String> res = new ArrayList<>();
      StringBuffer curVal = new StringBuffer();
      boolean quotes = false;
      boolean inval = false;

      while (ch >= 0)
      {
         if (quotes)
         {
            inval = true;
            if (ch == '\"')
            {
               quotes = false;
            }
            else
            {
               curVal.append((char) ch);
            }
         }
         else
         {
            if (ch == '\"')
            {
               quotes = true;
               if (inval)
               {
                  curVal.append('\"');
               }
            }
            else if (ch == ',')
            {
               res.add(curVal.toString());
               curVal = new StringBuffer();
               inval = false;
            }
            else if (ch == '\r')
            {
               // skip
            }
            else if (ch == '\n')
            {
               break;
            }
            else
            {
               curVal.append((char) ch);
            }
         }
         ch = r.read();
      }

      res.add(curVal.toString());
      return res;
   }

   /**
    * The method disables validation of SSL certificates when initiating HTTPS client connections
    * using java.net facility. The method creates new SSL Socket Factory and Host Name Verifier and
    * sets them as the defaults in {@linkplain HttpsURLConnection}.
    * <p>
    * All connections initiated with {@linkplain HttpsURLConnection} will be made without certificate
    * validations! Do not use on production systems!
    */
   public static void disableClientSSLCertificateValidation()
   {
      // do not validate certificate chains
      TrustManager[] trustManager = new TrustManager[]
      {
         new X509TrustManager()
         {
            @Override
            public X509Certificate[] getAcceptedIssuers()
            {
               return null;
            }

            @Override
            public void checkClientTrusted(X509Certificate[] chain, String authType)
            {
            }

            @Override
            public void checkServerTrusted(X509Certificate[] chain, String authType)
            {
            }
         }
      };

      try
      {
         // install socket factory accepting all certificates
         SSLContext sc = SSLContext.getInstance("SSL");
         sc.init(null, trustManager, new SecureRandom());
         HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());

         // accept all hosts
         HttpsURLConnection.setDefaultHostnameVerifier((hostname, session) -> true);

         LOG.log(Level.WARNING, "!!! Disabled SSL certificate validation for outgoing HTTPS connections. " +
                                "This is intended for test environments only !!! ");
      }
      catch (Exception ex)
      {
         LOG.log(Level.SEVERE, "Failure during HttpsURLConnection setup!", ex);
      }
   }

   /**
    * Poll all elements of a Queue and return them in the same order in a List. Useful when the elements 
    * have to be returned and the Queue cleared without losing any newly added elements. 
    * 
    * @param   queue
    *          A queue of T elements.
    * @param   <T>
    *          The type of elements in the Queue.
    *
    * @return  Empty Optional when the queue is empty or Optional of List with all the elements in the queue.
    */
   public static <T> Optional<List<T>> pollAll(Queue<T> queue)
   {
      if (queue == null || queue.isEmpty())
      {
         return Optional.empty();
      }
      List<T> elements = new ArrayList<>(queue.size());
      while (!queue.isEmpty())
      {
         elements.add(queue.poll());
      }
      return Optional.of(elements);
   }

   /**
    * Collect all the implemented interfaces (directly and from the hierarchy), for the given
    * class.
    * <p>
    * The collection is performed doing a BFS walk of the hierarchy, starting from the given class.
    * 
    * @param    cls
    *           The class from where to get the interfaces.
    * @param    interfaces
    *           A set where to store the interfaces, in the BFS order.
    * @param    collected
    *           A set of interfaces already processed.
    */
   private static void collectInterfacesImpl(Class<?> cls, Set<Class<?>> interfaces, Set<Class<?>> collected)
   {
      if (collected.contains(cls))
      {
         return;
      }
      
      Class<?>[] ifcs = cls.getInterfaces();
      if (ifcs != null)
      {
         for (Class<?> ifc : ifcs)
         {
            interfaces.add(ifc);
         }
         
         for (Class<?> ifc : ifcs)
         {
            collectInterfacesImpl(ifc, interfaces, collected);

            collected.add(ifc);
         }
      }

      if (cls.getSuperclass() != null)
      {
         collectInterfacesImpl(cls.getSuperclass(), interfaces, collected);
      }
   }

   /**
    * 
    * Utility class extending HashMap functionality
    *
    * @param <K> 
    *        class of the key
    * @param <V>
    *        class of the value
    */
   public static class InstrumentedMap<K,V> extends HashMap<K,V>
   {
      /** the value supplier */
      private Supplier<V> valueSupplier;
      
      /**
       * Constructor
       * 
       * @param valueSupplier
       *        the value supplier
       */
      public InstrumentedMap(Supplier<V> valueSupplier)
      {
         super();
         this.valueSupplier = valueSupplier;
      }

      /**
       * Modify the value of a map creating it if absent
       *  
       * @param key
       *        the key
       * @param modifier
       *        the value modifier
       *        
       * @return the value
       */
      public V modifyValue(K key, Consumer<V> modifier)
      {
         V value = get(key);
         if (value == null)
         {
            put(key, value = valueSupplier.get());
         }
         modifier.accept(value);
         return value;
      }
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Internal counter tha last used value */ 
      private int counter = -1;
      
      /**
       * Project token. This is used for per-project directory settings in case multiple projects
       * are 'parked' within the same P2J installation and setting are different.
       */
      private String projectToken = null;

      /** The name of the FWD client's temporary folder. */
      private String defaultFwdTempFolder = "fwd" + new Random().nextLong();
   }
   
}
