/*
 ** Module   : CentralLoggerFallback.java
 ** Abstract : CentralLogger concrete default / fallback version.
 **
 ** Copyright (c) 2023, Golden Code Development Corporation.
 **
 ** -#- -I- --Date-- ----------------------------Description-----------------------------
 ** 001 GBB 20230207 Initial setup
 ** 002 GBB 20230517 Default level support added to filter out lower level logs.
 ** 003 GBB 20230619 Adding field level to each logger instance to boost isLoggable performance.
 */
/*
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU Affero General Public License as
 ** published by the Free Software Foundation, either version 3 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU Affero General Public License for more details.
 **
 ** You may find a copy of the GNU Affero GPL version 3 at the following
 ** location: https://www.gnu.org/licenses/agpl-3.0.en.html
 **
 ** Additional terms under GNU Affero GPL version 3 section 7:
 **
 **   Under Section 7 of the GNU Affero GPL version 3, the following additional
 **   terms apply to the works covered under the License.  These additional terms
 **   are non-permissive additional terms allowed under Section 7 of the GNU
 **   Affero GPL version 3 and may not be removed by you.
 **
 **   0. Attribution Requirement.
 **
 **     You must preserve all legal notices or author attributions in the covered
 **     work or Appropriate Legal Notices displayed by works containing the covered
 **     work.  You may not remove from the covered work any author or developer
 **     credit already included within the covered work.
 **
 **   1. No License To Use Trademarks.
 **
 **     This license does not grant any license or rights to use the trademarks
 **     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
 **     of Golden Code Development Corporation. You are not authorized to use the
 **     name Golden Code, FWD, or the names of any author or contributor, for
 **     publicity purposes without written authorization.
 **
 **   2. No Misrepresentation of Affiliation.
 **
 **     You may not represent yourself as Golden Code Development Corporation or FWD.
 **
 **     You may not represent yourself for publicity purposes as associated with
 **     Golden Code Development Corporation, FWD, or any author or contributor to
 **     the covered work, without written authorization.
 **
 **   3. No Misrepresentation of Source or Origin.
 **
 **     You may not represent the covered work as solely your work.  All modified
 **     versions of the covered work must be marked in a reasonable way to make it
 **     clear that the modified work is not originating from Golden Code Development
 **     Corporation or FWD.  All modified versions must contain the notices of
 **     attribution required in this license.
 */
package src.com.goldencode.p2j.util.logging;

import java.util.*;
import java.util.logging.Formatter;
import java.util.logging.*;

import static src.com.goldencode.p2j.util.logging.CentralLogFormatter.*;

/** A concrete version of CentralLogger as fallback / a default implementation that logs to System streams. */
class CentralLoggerFallback
extends CentralLogger
{
   /** The formatter for logs. */
   private static final Formatter LOG_FORMATTER = new SimpleLogFormatter();

   /** The root logger's level. */
   private static Level rootLevel = DEFAULT_ROOT_LEVEL;
   
   static
   {
      CentralLogger.completePreInitWork(new HashMap<>(), rootLevel);
   }
   
   /**
    * Package-private constructor for CentralLoggerFallback.
    *
    * @param    loggerName
    *           The name of the logger.
    * @param    level
    *           The level of the logger.
    * @param    checkParentLevels
    *           Flag to prevent recursion for class loaders.
    * @param    excludeSMContext
    *           Flag to prevent deadlock on SecurityManager.
    */
   CentralLoggerFallback(String loggerName, Level level, boolean checkParentLevels, boolean excludeSMContext)
   {
      super(loggerName, level, checkParentLevels, excludeSMContext);
   }

   /**
    * Sets the root logger's level. Expected to be the first statement in the entry class main method.
    * 
    * @param    level
    *           The new root logger's level.
    */
   public static void setRootLevel(Level level)
   {
      rootLevel = level;
   }   
   
   /**
    * Publishes a log record by printing it to System.out. Log records with level WARNING and ERROR are 
    * printed on System.err.
    *
    * @param    logRecord
    *           A log record.
    */
   @Override
   protected void publish(LogRecord logRecord)
   {
      if (logRecord.getLevel().intValue() < rootLevel.intValue())
      {
         return;
      }
      if (logRecord.getLevel().intValue() >= Level.WARNING.intValue())
      {
         System.err.print(LOG_FORMATTER.format(logRecord));
         return;
      }
      System.out.print(LOG_FORMATTER.format(logRecord));
   }

   /** Formatter for log records. Prints the log level, the message and the throwable details. */
   private static class SimpleLogFormatter
   extends Formatter
   {
      /** The line separator on the system. */
      private static final String LINE_SEPARATOR = System.lineSeparator();

      /**
       * Returns the string representation of the log record. Reuses methods from the {@link CentralLogFormatter}.
       * 
       * @param    record
       *           A log record.
       *           
       * @return   The string representation of the log record.
       */
      @Override
      public String format(LogRecord record)
      {
         StringBuilder sb = new StringBuilder();
         
         dumpMessage(sb, record);

         if (record.getThrown() != null)
         {
            sb.append(LINE_SEPARATOR);
            dumpThrowable(sb, record.getThrown());
         }

         sb.append(LINE_SEPARATOR);

         return sb.toString();
      }
   }
   
}
