/*
 ** Module   : CentralLoggerNoModeWrapper.java
 ** Abstract : CentralLogger wrapper for the undetermined modes: class loaders, CLIENT_BASE and SERVER_BASE.
 **
 ** Copyright (c) 2023, Golden Code Development Corporation.
 **
 ** -#- -I- --Date-- ----------------------------Description-----------------------------
 ** 001 GBB 20230207 Initial setup
 ** 002 GBB 20230519 Resolve logger mode after server params are parsed and server mode is determined.
 ** 003 GBB 20230619 Adding field level to each logger instance to boost isLoggable performance.
 */
/*
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU Affero General Public License as
 ** published by the Free Software Foundation, either version 3 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU Affero General Public License for more details.
 **
 ** You may find a copy of the GNU Affero GPL version 3 at the following
 ** location: https://www.gnu.org/licenses/agpl-3.0.en.html
 **
 ** Additional terms under GNU Affero GPL version 3 section 7:
 **
 **   Under Section 7 of the GNU Affero GPL version 3, the following additional
 **   terms apply to the works covered under the License.  These additional terms
 **   are non-permissive additional terms allowed under Section 7 of the GNU
 **   Affero GPL version 3 and may not be removed by you.
 **
 **   0. Attribution Requirement.
 **
 **     You must preserve all legal notices or author attributions in the covered
 **     work or Appropriate Legal Notices displayed by works containing the covered
 **     work.  You may not remove from the covered work any author or developer
 **     credit already included within the covered work.
 **
 **   1. No License To Use Trademarks.
 **
 **     This license does not grant any license or rights to use the trademarks
 **     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
 **     of Golden Code Development Corporation. You are not authorized to use the
 **     name Golden Code, FWD, or the names of any author or contributor, for
 **     publicity purposes without written authorization.
 **
 **   2. No Misrepresentation of Affiliation.
 **
 **     You may not represent yourself as Golden Code Development Corporation or FWD.
 **
 **     You may not represent yourself for publicity purposes as associated with
 **     Golden Code Development Corporation, FWD, or any author or contributor to
 **     the covered work, without written authorization.
 **
 **   3. No Misrepresentation of Source or Origin.
 **
 **     You may not represent the covered work as solely your work.  All modified
 **     versions of the covered work must be marked in a reasonable way to make it
 **     clear that the modified work is not originating from Golden Code Development
 **     Corporation or FWD.  All modified versions must contain the notices of
 **     attribution required in this license.
 */
package src.com.goldencode.p2j.util.logging;

import java.util.logging.*;

/**
 * Proxy for other types of loggers. Used for classes loaded by a custom class loader before the app starts
 * its main method and sets logger mode. Used also for base server/client mode, where the concrete type is 
 * still not set by the configs/startup params.
 */
class CentralLoggerNoModeWrapper
extends CentralLogger
{
   /** The target logger, instantiated after the logger mode is determined. */
   private CentralLogger logger;
   
   /**
    * Package-private constructor for CentralLoggerForClassLoaders.
    *
    * @param    loggerName
    *           The name of the logger.
    * @param    level
    *           The level of the logger.
    * @param    checkParentLevels
    *           Flag to prevent recursion for class loaders.
    * @param    excludeSMContext
    *           Flag to prevent deadlock on SecurityManager.
    */
   CentralLoggerNoModeWrapper(String loggerName, Level level, boolean checkParentLevels, boolean excludeSMContext)
   {
      super(loggerName, level, checkParentLevels, excludeSMContext);
   }

   /**
    * Publishes a log record by adding it to the pre-init buffer before the initialization kicks in, or 
    * forwarding it to the target logger.
    * 
    * @param    logRecord
    *           The log record.
    */
   @Override
   protected void publish(LogRecord logRecord)
   {
      Mode mode = getMode();
      if (mode == null || mode == Mode.CLIENT_BASE || mode == Mode.SERVER_BASE)
      {
         addToPreInitBuffer(logRecord);
         return;
      }
      if (logger == null)
      {
         logger = CentralLogger.get(getLoggerName(), isCheckParentLevels(), isExcludeSMContext());
      }
      logger.publish(logRecord);
   }
}
