/*
** Module   : NativeTypeSerializer.java
** Abstract : Helper class to serialize native values, when the value's type is unknown at compile time.
**
** Copyright (c) 2014-2023, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------------Description---------------------------------------
** 001 CA  20141002 Created initial version.
** 002 IAS 20150702 Added cautious null processing.
** 003 IAS 20200721 Added helper methods for (de)serialization if int[] and Integer.
** 004 IAS 20200914 More helper methods added
**     IAS 20201006 Added readByte/writeByte
**     GES 20201023 Added read/writeSparseArray() methods.
**     CA  20201029 Fixed a bug in readSparseExtArray - readFully(byte[]) must be used.
**     VVT 20210104 readDouble(): javadoc fixed.
**     CA  20210317 Optimized the (de)serialization of byte arrays.
**     CA  20220106 Use writeObject for strings longer than 0xFFFF, as this is the limit for writeUTF.
**     TJD 20220504 Java 11 compatibility minor changes
** 005 GBB 20230512 Logging methods replaced by CentralLogger/ConversionStatus.
** 006 RAA 20230525 Creating a new BaseDataType instance is now done using BaseDataTypeFactory.
*/ 

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package src.com.goldencode.util;

import src.com.goldencode.p2j.util.*;
import src.com.goldencode.p2j.util.logging.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.function.*;

/**
 * Helper class to serialize native values, useful in cases when the value's type is unknown at
 * compile time (i.e. reflection is used to get values from a {@link Field}).
 * <p>
 * When using this class, all writing needs to be done using the {@link #writeValue} API and it
 * must be read using the {@link #readValueWorker} API.
 * <p>
 * All Java 7 native types are supported, plus {@link String}.  If the class passed to 
 * {@link #getWorker} is not a native type or {@link String}, then it will fallback to serializing
 * the value as a plain {@link Object}.
 */
public abstract class NativeTypeSerializer
{
   /** A serializer for native <code>byte</code> values. */
   public static final ByteSerializer BYTE_SERIALIZER = new ByteSerializer();
  
   /** A serializer for native <code>short</code> values. */
   public static final ShortSerializer SHORT_SERIALIZER = new ShortSerializer();
   
   /** A serializer for native <code>int</code> values. */
   public static final IntSerializer INT_SERIALIZER = new IntSerializer();
   
   /** A serializer for native <code>long</code> values. */
   public static final LongSerializer LONG_SERIALIZER = new LongSerializer();
   
   /** A serializer for native <code>float</code> values. */
   public static final FloatSerializer FLOAT_SERIALIZER = new FloatSerializer();
   
   /** A serializer for native <code>double</code> values. */
   public static final DoubleSerializer DOUBLE_SERIALIZER = new DoubleSerializer();
   
   /** A serializer for native <code>boolean</code> values. */
   public static final BooleanSerializer BOOLEAN_SERIALIZER = new BooleanSerializer();
   
   /** A serializer for native <code>char</code> values. */
   public static final CharSerializer CHAR_SERIALIZER = new CharSerializer();
   
   /** A serializer for {@link String} values. */
   public static final StringSerializer STRING_SERIALIZER = new StringSerializer();
   
   /** Fallback serializer for other {@link Object} instances. */
   public static final ObjectSerializer OBJECT_SERIALIZER = new ObjectSerializer();
   
   /** An array of workers, identified by their {@link #mark}. */
   private static final NativeTypeSerializer[] workers = new NativeTypeSerializer[]
   {
      BYTE_SERIALIZER,
      SHORT_SERIALIZER,
      INT_SERIALIZER,
      LONG_SERIALIZER,
      FLOAT_SERIALIZER,
      DOUBLE_SERIALIZER,
      BOOLEAN_SERIALIZER,
      CHAR_SERIALIZER,
      STRING_SERIALIZER,
      OBJECT_SERIALIZER
   };
   
   /** Logger */
   private static final CentralLogger LOG = CentralLogger.get(NativeTypeSerializer.class);

   /** A special mark to identify the type of the written value, when reading. */
   protected final byte mark;

   /**
    * Create a new serializer and set its mark.
    * 
    * @param    mark
    *           The serializer's mark.
    */
   protected NativeTypeSerializer(int mark)
   {
      this.mark = (byte) mark;
   }
   
   /**
    * Write the given value using the specified {@link NativeTypeSerializer worker}, to the 
    * specified destination.
    * <p>
    * The first written byte will be the worker's {@link #mark}, followed by the serialized value.
    * 
    * @param    value
    *           The value to be written.
    * @param    worker
    *           The worker to write the value.
    * @param    out
    *           The destination.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   public static void writeValue(Object value, NativeTypeSerializer worker, ObjectOutput out)
   throws IOException
   {
      if (value == null)
      {
         out.writeByte(OBJECT_SERIALIZER.mark);
         OBJECT_SERIALIZER.writeValueWorker(value, out);
         return;
      }
      out.writeByte(worker.mark);
      worker.writeValueWorker(value, out);
   }

   /**
    * Write integer array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeIntArray(ObjectOutput out, int[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         for (int i = 0; i < a.length; i++)
         {
            out.writeInt(a[i]);
         }
      }
   }
   
   /**
    * Read Integer array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Integer array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Integer[] readIntegerArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      Integer[] a = new Integer[len];
      for (int i = 0; i < len; i++)
      {
         a[i] = readInteger(in);
      }
      return a;
   }

   /**
    * Write Integer array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeIntegerArray(ObjectOutput out, Integer[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         for (int i = 0; i < a.length; i++)
         {
            writeInteger(out, a[i]);
         }
      }
   }

   /**
    * Read int array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return int array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static int[] readIntArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      int[] a = new int[len];
      for (int i = 0; i < len; i++)
      {
         a[i] = in.readInt();
      }
      return a;
   }
   
   /**
    * Read String array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return String array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static String[] readStringArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      String[] a = new String[len];
      for (int i = 0; i < len; i++)
      {
         a[i] = readString(in);
      }
      return a;
   }

   /**
    * Write String array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeStringArray(ObjectOutput out, String[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         for (int i = 0; i < a.length; i++)
         {
            writeString(out, a[i]);
         }
      }
   }
   
   /**
    * Write String set to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        set to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeStringSet(ObjectOutput out, Set<String> a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.size());
      if (a != null)
      {
         for (String s: a)
         {
            out.writeUTF(s);
         }
      }
   }
   
   /**
    * Read Integer set from the ObjectInput
    * 
    * @param in
    *        ObjectOutput instance
    * @return Integer set 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Set<Integer> readIntSet(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      Set<Integer> a = new HashSet<>();
      for (int i = 0; i < len; i++)
      {
         a.add(in.readInt());
      }
      return a;
   }
   
   /**
    * Write Long set to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        set to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeLongSet(ObjectOutput out, Set<Long> a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.size());
      if (a != null)
      {
         for (Long i: a)
         {
            out.writeLong(i.longValue());
         }
      }
   }
   
   /**
    * Read Long set from the ObjectInput
    * 
    * @param in
    *        ObjectOutput instance
    * @return Long set 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Set<Long> readLongSet(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      Set<Long> a = new HashSet<>();
      for (int i = 0; i < len; i++)
      {
         a.add(in.readLong());
      }
      return a;
   }
   
   /**
    * Write Integer set to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        set to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeIntSet(ObjectOutput out, Set<Integer> a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.size());
      if (a != null)
      {
         for (Integer i: a)
         {
            out.writeInt(i.intValue());
         }
      }
   }
   
   /**
    * Read String set from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @param notNull
    *        if <code>true</code> return emptySet instead of <code>null</code> 
    * 
    * @return String set 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Set<String> readStringSet(ObjectInput in, boolean notNull)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return notNull ? Collections.emptySet() : null;
      }
      Set<String> a = new HashSet<>();
      for (int i = 0; i < len; i++)
      {
         a.add(in.readUTF());
      }
      return a;
   }
   

   /**
    * Write Map to the ObjectOutput
    * @param <K> 
    *        key type
    * @param <V> 
    *        value type
    * @param out
    *        ObjectOutput instance
    * @param map
    *        array to be written
    * @param keyWriter
    *        writer for keys 
    * @param valWriter 
    *        writer for values
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <K,V> void writeMap(ObjectOutput out, Map<K,V> map, 
         Writer<K> keyWriter, Writer<V> valWriter) 
   throws IOException
   {
      out.writeInt( map == null ? -1 : map.size());
      if (map != null)
      {
         for (Map.Entry<K,V> e: map.entrySet())
         {
            keyWriter.write(out, e.getKey());
            valWriter.write(out, e.getValue());
         }
      }
   }
   
   /**
    * Read byte array from the ObjectInput
    * @param <M> 
    *        map type
    * @param <K> 
    *        key type
    * @param <V> 
    *        value type
    * @param in
    *        ObjectInput instance
    * @param allocator 
    *        Map allocator
    * @param keyReader
    *        reader for keys 
    * @param valReader 
    *        reader for values 
    * @return Map 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         In case when class is not found.
    */
   public static <M extends Map<K,V>, K,V> M readMap(ObjectInput in, Supplier<M> allocator,
         Reader<K> keyReader, Reader<V> valReader) 
   throws IOException, ClassNotFoundException
   {
      int size = in.readInt();
      if (size < 0)
      {
         return null;
      }
      M map = allocator.get();
      while (size-- > 0)
      {
         K key = keyReader.read(in);
         map.put(key, valReader.read(in));
      }
      return map;
   }

   /**
    * Write byte array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeByteArray(ObjectOutput out, byte[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         out.write(a);
      }
   }
   
   /**
    * Read byte array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return byte array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static byte[] readByteArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      byte[] a = new byte[len];
      in.readFully(a);

      return a;
   }
   
   /**
    * Write short array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeShortArray(ObjectOutput out, short[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         for (int i = 0; i < a.length; i++)
         {
            out.writeShort(a[i]);
         }
      }
   }
   
   /**
    * Read short array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return short array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static short[] readShortArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      short[] a = new short[len];
      for (int i = 0; i < len; i++)
      {
         a[i] = in.readShort();
      }
      return a;
   }
   
   /**
    * Write long array to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param a
    *        array to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeLongArray(ObjectOutput out, long[] a) 
   throws IOException
   {
      out.writeInt( a == null ? -1 : a.length);
      if (a != null)
      {
         for (int i = 0; i < a.length; i++)
         {
            out.writeLong(a[i]);
         }
      }
   }
   
   /**
    * Read short array from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return short array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static long[] readLongArray(ObjectInput in)
   throws IOException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      long[] a = new long[len];
      for (int i = 0; i < len; i++)
      {
         a[i] = in.readLong();
      }
      return a;
   }
   

   /**
    * Write Integer to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeInteger(ObjectOutput out, Integer value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         out.writeInt(value.intValue());
      }
   }
   
   /**
    * Read Integer from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Integer value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Integer readInteger(ObjectInput in)
   throws IOException
   {
      if (in.readByte() == 0)
      {
         return null;
      }
      return Integer.valueOf(in.readInt());
   }

   /**
    * Write Byte to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeByte(ObjectOutput out, Byte value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         out.writeByte(value.byteValue());
      }
   }
   
   /**
    * Read Byte from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Integer value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Byte readByte(ObjectInput in)
   throws IOException
   {
      if (in.readByte() == 0)
      {
         return null;
      }
      return Byte.valueOf(in.readByte());
   }

   /**
    * Write Longr to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeLong(ObjectOutput out, Long value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         out.writeLong(value.longValue());
      }
   }
   
   /**
    * Read Long from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Long value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Long readLong(ObjectInput in)
   throws IOException
   {
      if (in.readByte() == 0)
      {
         return null;
      }
      return Long.valueOf(in.readLong());
   }

   /**
    * Write Boolean to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeBoolean(ObjectOutput out, Boolean value)
   throws IOException
   {
      out.writeByte(value == null ? 0x7f : value.booleanValue() ? 1 : 0);
   }
   
   /**
    * Read Boolean from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Boolean value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Boolean readBoolean(ObjectInput in)
   throws IOException
   {
      byte b = in.readByte();
      return b == 0x7f ? null : b == 1 ? Boolean.TRUE : Boolean.FALSE;
   }

   /**
    * Write enum to the ObjectOutput
    * 
    * @param <T> 
    *        enum type
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T extends Enum<?>> void writeEnum(ObjectOutput out, T value)
   throws IOException
   {
      out.writeInt(value == null ? -1 : value.ordinal());
   }
   
   /**
    * Read enum from the ObjectInput
    * 
    * @param <T> 
    *        enum type
    * @param in
    *        ObjectInput instance
    * @param values
    *        enum <code>values()</code> array
    * @return enum value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T extends Enum<?>> T readEnum(ObjectInput in, T[] values)
   throws IOException
   {
      int ord = in.readInt();
      return ord < 0 ? null : values[ord];
   }

   /**
    * Write Double to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeDouble(ObjectOutput out, Double value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         out.writeDouble(value.doubleValue());
      }
   }
   
   /**
    * Read Double from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return Double value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static Double readDouble(ObjectInput in)
   throws IOException
   {
      if (in.readByte() == (byte)0)
      {
         return null;
      }
      return Double.valueOf(in.readDouble());
   }

   /**
    * Write a sparse array to the ObjectOutput, where {@code null} elements are skipped.
    * 
    * @param    <T> 
    *           The array element type.
    * @param    out
    *           ObjectOutput instance.
    * @param    value
    *           The value to be written.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   public static <T extends Externalizable> void writeSparseArray(ObjectOutput out, T[] value)
   throws IOException
   {
      // write overall array size
      out.writeInt(value == null ? -1 : value.length);
      
      if (value != null && value.length > 0)
      {
         BitSet bits = new BitSet(value.length);
         
         int nonNull = 0;
         
         // calculate how many non-null values exist
         for (int i = 0; i < value.length; i++)
         {
            if (value[i] != null)
            {
               nonNull++;
               bits.set(i);
            }
         }
         
         // send the number of non-null values that will follow
         out.writeInt(nonNull);
         
         // instead of sending each index as an integer, we send a bitset which marks valid index positions
         byte[] bytes = bits.toByteArray();
         out.writeInt(bytes.length);
         out.write(bytes);
         
         for (int j = 0; j < value.length; j++)
         {
            // write out each non-null as (index, value)
            if (value[j] != null)
            {
               writeExternalizable(out, value[j]);
            }
         }
      }
   }
   
   /**
    * Write array to the ObjectOutput
    * @param <T> 
    *        array element type
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T extends Externalizable> void writeArray(ObjectOutput out, T[] value)
   throws IOException
   {
      out.writeInt(value == null ? -1 : value.length);
      if (value != null)
      {
         for (T val: value)
         {
            writeExternalizable(out, val);
         }
      }
   }
   
   /**
    * Read a sparse array from the ObjectInput, where {@code null} elements are skipped.
    * 
    * @param    <T> 
    *           The array element type.
    * @param    in
    *           ObjectInput instance.
    * @param    allocate
    *           The array allocator.
    * @param    init
    *           The element default constructor.
    *           
    * @return   The array. 
    * 
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException 
    *           If the class is not found.
    */
   public static <T extends Externalizable> T[] readSparseExtArray(ObjectInput      in, 
                                                                   IntFunction<T[]> allocate, 
                                                                   Supplier<T>      init)
   throws IOException,
          ClassNotFoundException
   {
      int len = in.readInt();
      
      if (len < 0)
      {
         return null;
      }
      
      T[] a = allocate.apply(len);
      
      if (len > 0)
      {
         int nonNull = in.readInt();
         
         // read in the bitset
         int    numBytes = in.readInt();
         byte[] bytes    = new byte[numBytes];
         
         in.readFully(bytes);
         BitSet bits = BitSet.valueOf(bytes);

         int idx = 0;
         
         for (int i = 0; i < nonNull; i++)
         {
            // use the bitset to calculate the next index position
            idx = bits.nextSetBit(idx);
            
            a[idx] = readExternalizable(in, init);
            
            idx++;
         }      
      }
      
      return a;
   }

   /**
    * Read array from the ObjectInput
    * 
    * @param <T> 
    *        array element type
    * @param in
    *        ObjectInput instance
    * @param  allocate
    *         array allocator
    * @param  init
    *         default constructor
    * @return array 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If class is not found
    */
   public static <T extends Externalizable> T[] readExtArray(ObjectInput in, 
         IntFunction<T[]> allocate, Supplier<T> init)
   throws IOException, ClassNotFoundException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      T[] a = allocate.apply(len);
      for (int i = 0; i < len; i++)
      {
         a[i] = readExternalizable(in, init);
      }
      return a;
   }

   /**
    * Write list to the ObjectOutput
    * @param <T> 
    *        list element type
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T extends Externalizable> void writeList(ObjectOutput out, List<T> value)
   throws IOException
   {
      out.writeInt(value == null ? -1 : value.size());
      if (value != null)
      {
         for (T val: value)
         {
            writeExternalizable(out, val);
         }
      }
   }
   
   /**
    * Write list to the ObjectOutput
    * @param <T> 
    *        list element type
    * @param out
    *        ObjectOutput instance
    * @param writer
    *        writer
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T> void writeList(ObjectOutput out, Writer<T> writer, List<T> value)
   throws IOException
   {
      out.writeInt(value == null ? -1 : value.size());
      if (value != null)
      {
         for (T val: value)
         {
            writer.write(out, val);
         }
      }
   }

   /**
    * Read List from the ObjectInput
    * 
    * @param <T> 
    *        list element type

    * @param in
    *        ObjectInput instancelocator
    * @param  init
    *         default constructor
    * @return List 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException
    *         If type is not found 
    */
   public static <T extends Externalizable> List<T> readList(ObjectInput in, Supplier<T> init)
   throws IOException, ClassNotFoundException
   {
      return readList(in, init, false);
   }

   /**
    * Read List from the ObjectInput
    * 
    * @param <T> 
    *        list element type
    * @param in
    *        ObjectInput instancelocator
    * @param  init
    *         default constructor
    * @param notNull
    *        if <code>true</code> return emptyList instead of <code>null</code> 
    * @return List 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If type is not found 
    */
   public static <T extends Externalizable> List<T> readList(ObjectInput in, 
         Supplier<T> init, boolean notNull)
   throws IOException, ClassNotFoundException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return notNull ? Collections.emptyList() : null;
      }
      List<T> a = new ArrayList<T>(len);
      for (int i = 0; i < len; i++)
      {
         a.add(readExternalizable(in, init));
      }
      return a;
   }

   /**
    * Read List from the ObjectInput
    * 
    * @param <T> 
    *        list element type
    * @param in
    *        ObjectInput instancelocator
    * @param  reader
    *         default constructor
    * @return List value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If type is not found 
    */
   public static <T> List<T> readList(ObjectInput in, Reader<T> reader)
   throws IOException, ClassNotFoundException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      List<T> a = new ArrayList<T>(len);
      for (int i = 0; i < len; i++)
      {
         T e = reader.read(in);
         a.add(e);
      }
      return a;
   }

   /**
    * Read List subclass from the ObjectInput
    * @param <L> 
    *        List type
    * @param <T> 
    *        elements' type
    * @param in
    *        ObjectInput instance locator
    * @param allocator 
    *         list allocator
    * @param  reader
    *         default constructor
    * @return List value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If type is not found 
    */
   public static <L extends List<T>, T> L readList(ObjectInput in, IntFunction<L> allocator, 
         Reader<T> reader)
   throws IOException, ClassNotFoundException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      L a = allocator.apply(len);
      for (int i = 0; i < len; i++)
      {
         T e = reader.read(in);
         a.add(e);
      }
      return a;
   }

   /**
    * Read array from the ObjectInput
    * 
    * @param <T> 
    *        array element type
    * @param in
    *        ObjectInput instancelocator
    * @param  allocate
    *         array allocator
    * @param  reader
    *         default constructor
    * @return List value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If type is not found 
    */
   public static <T> T[] readArray(ObjectInput in, IntFunction<T[]> allocate, Reader<T> reader)
   throws IOException, ClassNotFoundException
   {
      int len = in.readInt();
      if (len < 0)
      {
         return null;
      }
      T[] a = allocate.apply(len);
      for (int i = 0; i < len; i++)
      {
         T e = reader.read(in);
         a[i] = e;
      }
      return a;
   }

   /**
    * Write array to the ObjectOutput
    * @param <T> 
    *        array element type
    * @param out
    *        ObjectOutput instance
    * @param writer
    *        writer
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static <T> void writeArray(ObjectOutput out, Writer<T> writer, T[] value)
   throws IOException
   {
      out.writeInt(value == null ? -1 : value.length);
      if (value != null)
      {
         for (T val: value)
         {
            writer.write(out, val);
         }
      }
   }

  /**
    * Write Externalizable to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeExternalizable(ObjectOutput out, Externalizable value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         value.writeExternal(out);
      }
   }

   /**
    * Read Externalizable from the ObjectInput
    * @param <T> 
    *        value type
    * @param in
    *        ObjectInput instance
    * @param f
    *        object constructor
    * 
    * @return Externalizable value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If class is not found 
    */
   public static <T extends Externalizable> T readExternalizable(ObjectInput in, Supplier<T> f)
   throws IOException, ClassNotFoundException
   {
      if (in.readByte() == 0)
      {
         return null;
      }
      return readExtNotNull(in, f);
   }

   /**
    * Read not null Externalizable from the ObjectInput
    * @param <T> 
    *        value type
    * @param in
    *        ObjectInput instance
    * @param f
    *        object constructor
    * 
    * @return Externalizable value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    * @throws ClassNotFoundException 
    *         If class is not found 
    */
   public static <T extends Externalizable> T readExtNotNull(ObjectInput in, Supplier<T> f)
   throws IOException, ClassNotFoundException
   {
      T obj = f.get();
      obj.readExternal(in);
      return obj;
   }

   /**
    * Write String to the ObjectOutput
    * @param out
    *        ObjectOutput instance
    * @param value
    *        value to be written
    * @throws IOException
    *         In case of I/O errors.
    */
   public static void writeString(ObjectOutput out, String value)
   throws IOException
   {
      out.writeByte(value == null ? 0 : 1);
      if (value != null)
      {
         boolean asObject = value.length() > 0xFFFF; // maximum length is 65535 for writeUTF
         out.writeByte(asObject ? 1 : 0);
         if (asObject)
         {
            out.writeObject(value);
         }
         else
         {
            out.writeUTF(value);
         }
      }
   }
   
   /**
    * Read String from the ObjectInput
    * 
    * @param in
    *        ObjectInput instance
    * @return String value 
    * 
    * @throws IOException
    *         In case of I/O errors.
    */
   public static String readString(ObjectInput in)
   throws IOException
   {
      if (in.readByte() == 0)
      {
         return null;
      }
      
      boolean asObject = in.readByte() == 1;
      
      try
      {
         return asObject ? (String) in.readObject() : in.readUTF();
      } 
      catch (ClassNotFoundException e)
      {
         throw new IOException(e);
      }
   }

   /**
    * Helper class for (de)serialization classes with subclasses
    *
    * @param <T>
    *        base type
    */
   public static class Externalizer<T extends Externalizable>
   {
      private final Class<? extends T> base;
      private final Class<? extends T>[] subclasses;

      /**
       * Constructor
       * @param subclasses
       *        possible subclasses of the base class
       */
      public Externalizer(Class<? extends T>... subclasses)
      {
         this.subclasses = subclasses;
         this.base = subclasses[0];
      }
      
      /**
       * Write instance of the base class subclass to the output stream
       * @param out
       *        output stream
       * @param val
       *        value to be written
       * @throws IOException 
       *         In case of I/O error
       */
      public void write(ObjectOutput out, T val) throws IOException
      {
         if (val == null)
         {
            out.writeByte(0);
            return;
         }
         
         out.writeByte(type(val));val.writeExternal(out);
      }

      

   /**
    * Read instance of the base class subclass from the input stream
    * @param in
    *        input stream
    * @return
    *        instance of the base class subclass
    * @throws IOException
    *         In case of I/O error
    * @throws ClassNotFoundException 
    *         If class is not found 
    */
     public T read(ObjectInput in) throws IOException, ClassNotFoundException
     {
        int type = in.readByte();
        if (type == 0)
        {
           return null;
        }
        type--;
        if (type < 0 || type >= subclasses.length)
        {
           throw new IllegalStateException(
                 String.format("Unexpected subtype of %s: %s", base.getName(), type));
        }
        Class<? extends T> cls = subclasses[type];
        T val;
        try
        {
           val = cls.getDeclaredConstructor().newInstance();
        } catch (ReflectiveOperationException    |
                 IllegalArgumentException  |
                 SecurityException e)
        {
           throw new IllegalStateException("Cannot instantiate: " + cls.getName(), e);
        }
        val.readExternal(in);
        return val;
     }
     
     /**
      * Get type code of the value
      * @param val
      *        value
      * @return type code
      */
     private byte type(T val)
     {
        Class<?> vcls = val.getClass();
        int type = -1;
        for (int i = 0; i < subclasses.length; i++)
        {
           if (subclasses[i].equals(vcls))
           {
              type = i;
              break;
           }
        }
        if (type < 0)
        {
           throw new IllegalArgumentException(
                 String.format("Unexpected subclass of %s: %s", base.getName(), vcls.getName()));
        }
        return (byte) (type + 1);
     }
   }

//   public static void main(String... args) throws Exception
//   {
//      Externalizer<Color> ext = new Externalizer<>(Color.class, ColorGui.class);
//      LOG.info(ext.type(new Color()));
//      LOG.info(ext.type(new ColorGui()));
//      ext = new Externalizer<>(ColorGui.class, Color.class);
//      LOG.info(ext.type(new Color()));
//      LOG.info(ext.type(new ColorGui()));
//   }
   /**
    * Read the given value using the specified {@link NativeTypeSerializer worker}, from the 
    * specified destination.
    * <p>
    * The first read byte will be the worker's {@link #mark}, followed by the serialized value.
    * 
    * @param    in
    *           The source.
    * 
    * @return   The read value.
    * 
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If the worker is {@link ObjectSerializer} and the object's class can not be found.
    */
   public static Object readValue(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      byte worker = in.readByte();
      NativeTypeSerializer cfg = workers[worker];
      return cfg.readValueWorker(in);
   }

   /**
    * Get the {@link NativeTypeSerializer} associated with the given {@link Class}.
    * 
    * @param    ftype
    *           The class.
    *           
    * @return   See above.
    */
   public static NativeTypeSerializer getWorker(Class<?> ftype)
   {
      if (ftype == byte.class)
      {
         return BYTE_SERIALIZER;
      }
      else if (ftype == short.class)
      {
         return SHORT_SERIALIZER;
      }
      else if (ftype == int.class)
      {
         return INT_SERIALIZER;
      }
      else if (ftype == long.class)
      {
         return LONG_SERIALIZER;
      }
      else if (ftype == float.class)
      {
         return FLOAT_SERIALIZER;
      }
      else if (ftype == double.class)
      {
         return DOUBLE_SERIALIZER;
      }
      else if (ftype == boolean.class)
      {
         return BOOLEAN_SERIALIZER;
      }
      else if (ftype == char.class)
      {
         return CHAR_SERIALIZER;
      }
      else if (ftype == String.class)
      {
         return STRING_SERIALIZER;
      }
      else
      {
         return OBJECT_SERIALIZER;
      }
   }
   
   /**
    * A worker to write the value to the specified destination.
    * 
    * @param    value
    *           The value to be written.
    * @param    out
    *           The destination.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   protected abstract void writeValueWorker(Object value, ObjectOutput out)
   throws IOException;

   /**
    * A worker to read the value from the specified source.
    * 
    * @param    in
    *           The source.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If the worker is {@link ObjectSerializer} and the object's class can not be found.
    * @return   value read from the input
    */
   protected abstract Object readValueWorker(ObjectInput in)
   throws IOException,
          ClassNotFoundException;

   /**
    * A serializer for native <code>byte</code> values.
    */
   private static class ByteSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>byte</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>0</code>.
       */
      protected ByteSerializer()
      {
         super(0);
      }
      
      /**
       * Worker to write the value to the specified destination, as a native <code>byte</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeByte((byte) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>byte</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Byte readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readByte();
      }
   }
   
   /**
    * A serializer for native <code>short</code> values.
    */
   private static class ShortSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>short</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>1</code>.
       */
      protected ShortSerializer()
      {
         super(1);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>short</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeShort((short) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>short</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Short readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readShort();
      }
   }
   
   /**
    * A serializer for native <code>int</code> values.
    */
   private static class IntSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>int</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>2</code>.
       */
      protected IntSerializer()
      {
         super(2);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>int</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeInt((int) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>int</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Integer readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readInt();
      }
   }
   
   /**
    * A serializer for native <code>long</code> values.
    */
   private static class LongSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>long</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>3</code>.
       */
      protected LongSerializer()
      {
         super(3);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>long</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeLong((long) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>long</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Long readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readLong();
      }
   }
   
   /**
    * A serializer for native <code>float</code> values.
    */
   private static class FloatSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>float</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>4</code>.
       */
      protected FloatSerializer()
      {
         super(4);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>float</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeFloat((float) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>float</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Float readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readFloat();
      }
   }
   
   /**
    * A serializer for native <code>double</code> values.
    */
   private static class DoubleSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>double</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>5</code>.
       */
      protected DoubleSerializer()
      {
         super(5);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>double</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeDouble((double) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>double</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Double readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readDouble();
      }
   }
   
   /**
    * A serializer for native <code>boolean</code> values.
    */
   private static class BooleanSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>boolean</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>6</code>.
       */
      protected BooleanSerializer()
      {
         super(6);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>boolean</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeBoolean((boolean) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>boolean</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Boolean readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readBoolean();
      }
   }
   
   /**
    * A serializer for native <code>char</code> values.
    */
   private static class CharSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>char</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>7</code>.
       */
      protected CharSerializer()
      {
         super(7);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>char</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeChar((char) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>char</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Character readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readChar();
      }
   }
   
   /**
    * A serializer for {@link String} values.
    */
   private static class StringSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for {@link String} values and set its {@link NativeTypeSerializer#mark}
       * to <code>8</code>.
       */
      protected StringSerializer()
      {
         super(8);
      }

      /**
       * Worker to write the value to the specified destination, as a {@link String} value.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeUTF((String) value);
      }

      /**
       * A worker to read the value from the specified source, as a {@link String} value.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected String readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readUTF();
      }
   }
   
   /**
    * A fallback serializer for {@link Object} instances.
    */
   private static class ObjectSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for {@link Object} instances and set its 
       * {@link NativeTypeSerializer#mark} to <code>9</code>.
       */
      protected ObjectSerializer()
      {
         super(9);
      }

      /**
       * Worker to write the value to the specified destination, as a {@link Object} instance.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeObject(value);
      }

      /**
       * A worker to read the value from the specified source, as a {@link Object} instance.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       * @throws   ClassNotFoundException
       *           If the object's class can not be found.
       */
      @Override
      protected Object readValueWorker(ObjectInput in)
      throws IOException,
             ClassNotFoundException
      {
         return in.readObject();
      }
   }

   /**
    * Generic reader
    * @param <T>
    *        type of the objects to be read 
    */
   @FunctionalInterface
   public static interface Reader<T>
   {
      /**
       * Read object
       * @param in
       *        input stream
       * @return read object
       *    
       * @throws IOException
       *         In case of I/O error
       * @throws ClassNotFoundException
       *         if class is not found 
       */
      public T read(ObjectInput in) throws IOException, ClassNotFoundException;
   }

   /**
    * Generic reader
    * @param <T>
    *        type of the objects to be read 
    */
   @FunctionalInterface
   public static interface Writer<T>
   {
      /**
       * Write object
       * @param out
       *        output stream
       * @param value
       *        value to be written
       * @throws IOException
       *         In case of I/O error
       */
      public void write(ObjectOutput out, T value) throws IOException;
   }
}
