/*
** Module   : StringHelper.java
** Abstract : provides a set of useful helper methods
**
** Copyright (c) 2004-2023, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------------Description-----------------------------------
** 001 GES 20041109   @18684 Built as a subset of some utility functions
**                           written for the GCD Storefront application.
**                           This provides string formatting and
**                           manipulation functions.  These were
**                           written and tested by ECF but integrated
**                           (and in somecases documented) by GES.
** 002 GES 20041118   @18691 Added safeTrimLeading and safeTrimTailing
**                           methods.  These delete leading or trailing
**                           space characters from a String.
** 003 NVS 20050221   @19893 Added isRepeatedChar method, which verifies
**                           whether the specified string buffer is made 
**                           of copies of a single character.
** 004 ECF 20050304   @20172 Added sprintf method. Works the same way as
**                           the C runtime library function of the same
**                           name, except that the variable argument list
**                           is handled as an array of objects.
** 005 GES 20050321   @20450 Added a processEscapes() method to convert
**                           a string with escape sequences into a string
**                           with the proper resulting characters.
** 006 GES 20050421   @20625 Fixed sprintf() to handle embedded % chars
**                           in replacement strings.
** 007 ECF 20050412   @20702 Added stripEnclosing method.
** 008 GES 20050426   @20870 Changed processEscapes() method to fix a
**                           bug in unicode char parsing and to add octal
**                           escape sequence support.
** 009 NVS 20050510   @21448 Added countChar method which returns number
**                           of occurences of a char in a string.
** 010 GES 20050601   @21364 Updated sprintf() to handle all primitive
**                           types AND to convert all other Objects using
**                           toString() allowing any object to be passed
**                           as an argument as long as one uses %s as the
**                           specifier for such generic/unrecognized
**                           Objects.  The primitive types all match up
**                           with the specifiers as expected.
** 011 ECF 20050603   @21404 Modified 010 (@21364) to use String.valueOf()
**                           instead of toString(). This safely handles
**                           null references.
** 012 GES 20050617   @21508 Added method sweep() to replace tokens in
**                           a target string based on a map of key and
**                           value replacements.
** 013 ECF 20050628   @21602 More robust replacement algorithm inside
**                           sweep method. Escapes backslash and dollar
**                           sign in replacement value. Also added
**                           exception handler to report more useful
**                           diagnostics.
** 014 ECF 20051013   @23029 Added changeCase() method. Capitalizes or
**                           decapitalizes the first letter of a name.
** 015 GES 20051028   @23140 Added protection logic to changeCase().
** 016 ECF 20060113   @23914 Moved generateBeanName() from NameConverter.
**                           Method is needed at runtime; NameConverter is
**                           a conversion-time class only.
** 017 ECF 20060712   @27996 Changed safeTrimTrailing() to behave more
**                           like Progress. In addition to spaces, it now
**                           removes tabs, line feeds, and carriage
**                           returns as well.
** 018 GES 20060720   @28171 Added methods to render primative arrays to
**                           a text form.
** 019 GES 20060821   @28694 Added an alternate word wrap helper.
** 020 GES 20060925   @29868 Added removeChar().
** 021 ECF 20080305   @37464 Made slight optimizations to sweep() method.
** 022 CA  20090408   @41734 Fixed edge case for nolossWrapText(...) - 
**                           remove any white space after each line is 
**                           extracted.
** 023 CA  20090413   @41771 Javadoc change in nolossWrapText.
** 024 GES 20090422   @41919 Removed sprintf() method as all callers now use the
**                           standard formatting support.
** 025 GES 20090424   @41977 Package change.
** 026 EVL 20121102          Adding helper functions to double backslashes in
**                           string under Windows environment. Adding helper
**                           functions to substitute the filename in Windows
**                           environment if it contains the Linux file
**                           separator. The char '/' replaced with '\' if
**                           the current OS is Windows.
** 027 EVL 20130108          Shifting the Windows detection code to separate
**                           method in src.com.goldencode.util.PlatformHelper
**                           class. Call this static to detect the OS here.
** 028 VMN 20130411          Helper function parseTokens(...) is added
** 029 GES 20140228          Changed word wrapping to honor embedded newlines.
** 030 ECF 20140301          Minor cleanup of parseTokens method.
** 031 GES 20150124          Fixed octal escape sequence processing which was profoundly broken.
** 032 ECF 20150715          Replace StringBuffer with StringBuilder.
** 033 IAS 20150926          Added equals() method.
** 034 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 035 ECF 20160409          Rewrote sweep, which was horribly slow.
** 036 GES 20160309          Added a variant of stripEnclosing().
** 037 OM  20160603          Fixed rtrimming to space character only.
** 038 GES 20170302          Added stripMatchedComments().
** 039 SBI 20180209          Added getWordWithRightSpacesAt(content, offset).
** 040 SBI 20180731          Added getAverageWidths to calculate the average string widths.
** 041 SBI 20180829          Added getProportionalWidths to calculate substring widths.
** 042 SBI 20180913          Fixed getProportionalWidths if a non empty string has a zero width.
** 043 HC  20181003          Added ifNullOther method.
** 044 HC  20181124          Implemented TREEVIEW widget and related changes.
** 045 SBI 20190408          Fixed getWordWithRightSpacesAt(content, offset) to return a valid pair.
** 046 AIL 20190823          Fixed a bug where the last character was not selected properly by
**                           getWordWithRightSpacesAt().
**     AIL 20191001          Fixed a bug where getWordWithRightSpacesAt() may receive an invalid 
**                           offset.
** 047 GES 20200206          Added containsWhitespace().
** 048 CA  20200503          Added safeTrimTrailing(character) (used by CASE "literal").
** 049 GES 20210707          Added readStringList().
**     GES 20220323          Removed the file-system behavior from readStringList().
**     CA  20220412          Added convertToRegEx and readQuotedString.
** 050 HC  20230118          Added startsWithIgnoreCase and hashCodeCaseInsensitive.
** 051 CA  20230215          Cache the UNICODE uppercase chars, to avoid calling Character.toUpperCase, in
**                           hashCodeCaseInsensitive.
** 052 GBB 20230411          Added hasContent(String) method.
** 053 AL2 20230512          Extended hashCodeCaseInsensitive to account for right trimming.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package src.com.goldencode.util;

/**
 * Provides a set of methods to help in processing and formatting
 * <code>String</code> objects.  This is intended to augment the limited
 * set of formatting and processing methods in the <code>String</code> 
 * class.  Methods are provided to left pad, right pad, center, trim and
 * truncate among other features.
 * <p>
 * Most (if not all) of these methods are stateless and thus are implemented
 * as static methods. 
 */
public class StringHelper
{
   /**
    * Checks if the string has content. <code>true</code> if not null and not empty.
    *
    * @param   text
    *          The string to check for content.
    *
    * @return  Boolean to indicate if the content has content or not.
    */
   public static boolean hasContent(String text)
   {
      return text != null && !text.trim().isEmpty();
   }

}
